/* ----------------------------------------------------------------------- */
/*  AffyLib.c : implementation of the affymetrix library                   */
/*  Author : Ji HongKai ; Time: 2004.08                                    */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"

#include "MathLib.h"
#include "MatrixLib.h"
#include "StringLib.h"
#include "SequenceLib.h"
#include "GenomeLib.h"
#include "AffyLib.h"

/* ----------------------------------------------------------------------- */ 
/*                            Global Variables                             */
/* ----------------------------------------------------------------------- */ 
int AFFY_IS_BIG_ENDIAN = 1;

/* ----------------------------------------------------------------------- */ 
/*  Affy_CSVANNOT_To_GenomeAlign_200506()                                  */
/*  convert affy csv annotation file to genome mapping coordinates.        */
/* ----------------------------------------------------------------------- */ 
int Affy_CSVANNOT_To_GenomeAlign_200506(char strInFile[], char strOutFile[], char strSpecies[])
{
	/* define */
	FILE *fpIn;
	FILE *fpOut;

	char strLine[LONG_LINE_LENGTH];
	char strTemp[MEDLONG_LINE_LENGTH];
	char strChrPos[MEDLONG_LINE_LENGTH];
	char strRefLine[MEDLONG_LINE_LENGTH];
	char *chi,*chs;
	char *chSep,*chp;
	char strProbe[LINE_LENGTH];
	char strGenSym[LINE_LENGTH];
	char strUniGene[LINE_LENGTH];
	char strLocusID[LINE_LENGTH];
	int nLocusID;

	struct tagAffyGenomeAlign *vAlign;
	struct tagAffyGenomeAlign *vTrans;
	struct tagAffyGenomeAlign *pNewAlign,*pPrevAlign;
	struct tagString *vRefId[1024];
	int nRefCount;
	int ni,nLen,nLineId;
	
	
	/* init */
	vAlign = NULL;
	vTrans = NULL;
	fpIn = NULL;
	fpOut = NULL;

	/* open file */
	fpIn = fopen(strInFile,"r");
	if(fpIn == NULL)
	{
		printf("Error: Affy_CSVANNOT_To_GenomeAlign_200506, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = fopen(strOutFile,"w");
	if(fpOut == NULL)
	{
		fclose(fpIn);
		printf("Error: Affy_CSVANNOT_To_GenomeAlign_200506, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* read */
	nLineId = 0;
	fgets(strLine, LONG_LINE_LENGTH, fpIn);
	while(fgets(strLine, LONG_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		nLineId++;

		/* parse */
		vAlign = NULL;
		vTrans = NULL;

		/* 1: probe */
		if(strchr(strLine, '\"') == strLine)
		{
			chs = strLine+1;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chs = strLine;
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		strcpy(strProbe, chs);
		chs = chi+1;
		/* printf("%s\n", strProbe); */
		

		/* 2: GeneChip Array */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 3: Species Scientific Name */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 4: Annotation Date */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 5: Sequence Type */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 6: Sequence Source */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 7: Transcript ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 8: Target Description */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;
		

		/* 9: Representative Public ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 10-a: Archival UniGene Cluster */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		
		/* 10-b: UniGene ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		strcpy(strUniGene, chs);
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 11: Genome Version */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 12: Alignments */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}		
		strcpy(strChrPos, chs);
		chs = chi+1;

		if(strcmp(strChrPos, "---") == 0)
		{	
		}
		else
		{
			chp = strChrPos;
			chSep = strstr(chp, "///");
			while(chSep != NULL)
			{
				*chSep = '\0';
				strcpy(strTemp, chp);

				pNewAlign = NULL;
				pNewAlign = AffyGenomeAlignCreate();
				AffyGenomeAlignInit(pNewAlign, strTemp, strSpecies);
				if(vAlign == NULL)
				{
					vAlign = pNewAlign;
					pPrevAlign = pNewAlign;
				}
				else
				{
					pPrevAlign->pNext = pNewAlign;
					pPrevAlign = pNewAlign;
				}

				chp = chSep+3;
				chSep = strstr(chp, "///");
			}


			strcpy(strTemp, chp);
			pNewAlign = NULL;
			pNewAlign = AffyGenomeAlignCreate();
			AffyGenomeAlignInit(pNewAlign, strTemp, strSpecies);
			if(vAlign == NULL)
			{
				vAlign = pNewAlign;
				pPrevAlign = pNewAlign;
			}
			else
			{
				pPrevAlign->pNext = pNewAlign;
				pPrevAlign = pNewAlign;
			}
		}
			

		/* 13: Overlapping Transcripts */
		/* StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strChrPos, chs);

		chs = chi+1;

		if(strcmp(strChrPos, "---") == 0)
		{	
		}
		else
		{
			chp = strChrPos;
			chSep = strstr(chp, "///");
			while(chSep != NULL)
			{
				*chSep = '\0';
				strcpy(strTemp, chp);

				pNewAlign = NULL;
				pNewAlign = AffyGenomeAlignCreate();
				AffyGenomeTransInit(pNewAlign, strTemp, strSpecies);
				if(vTrans == NULL)
				{
					vTrans = pNewAlign;
					pPrevAlign = pNewAlign;
				}
				else
				{
					pPrevAlign->pNext = pNewAlign;
					pPrevAlign = pNewAlign;
				}

				chp = chSep+3;
				chSep = strstr(chp, "///");
			}


			strcpy(strTemp, chp);
			pNewAlign = NULL;
			pNewAlign = AffyGenomeAlignCreate();
			AffyGenomeTransInit(pNewAlign, strTemp, strSpecies);
			if(vTrans == NULL)
			{
				vTrans = pNewAlign;
				pPrevAlign = pNewAlign;
			}
			else
			{
				pPrevAlign->pNext = pNewAlign;
				pPrevAlign = pNewAlign;
			}
		}
		*/

		/* 14: Gene Title */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 15: Gene Symbol */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strGenSym, chs);
		chs = chi+1;

		if(strcmp(strGenSym, "---") == 0)
			strcpy(strGenSym, "NA");

		/* 16: Chromosomal Location */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 17: UniGene ClusterType */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		/* strcpy(strUniGene, chs); */
		chs = chi+1;

		/* 18: Ensembl */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 19: LocusLink or Entrez Gene */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strLocusID, chs);
		chs = chi+1;

		if(strcmp(strLocusID, "---") == 0)
		{
			nLocusID = -1;
		}
		else if(strstr(strLocusID, "///") != NULL)
		{
			nLocusID = -1;
		}
		else
		{
			nLocusID = atoi(strLocusID);
		}

		/* 20: SwissProt */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 21: EC */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 22: OMIM */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 23: RefSeq Protein ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 24: RefSeq Transcript ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strRefLine, chs);
		chs = chi+1;

		nRefCount = 0;
		ni = 0;
		if(strcmp(strRefLine, "---") == 0)
		{	
		}
		else
		{
			chp = strRefLine;
			chSep = strstr(chp, "///");
			while(chSep != NULL)
			{
				*chSep = '\0';
				strcpy(strTemp, chp);
				StrTrimRight(strTemp);
				StrTrimLeft(strTemp);
				nLen = strlen(strTemp)+1;
				vRefId[ni] = CreateString(nLen);
				strcpy(vRefId[ni]->m_pString, strTemp);
				ni++;

				chp = chSep+3;
				chSep = strstr(chp, "///");
			}

			strcpy(strTemp, chp);
			StrTrimRight(strTemp);
			StrTrimLeft(strTemp);
			nLen = strlen(strTemp)+1;
			vRefId[ni] = CreateString(nLen);
			strcpy(vRefId[ni]->m_pString, strTemp);
			ni++;

			nRefCount = ni;
		}

		
		/* 25: FlyBase */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 26: AGI */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 27: WormBase */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 28: MGI Name */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 29: RGD Name */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 30: SGD accession number */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 31: Gene Ontology Biological Process */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 32: Gene Ontology Cellular Component */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 33: Gene Ontology Molecular Function */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 34: Pathway */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 35: Protein Families */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 36: Protein Domains */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 37: InterPro */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 38: Trans Membrane */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 39: QTL */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 40: Annotation Description */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;
		
		/* 41: Annotation Transcript Cluster */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 42: Transcript Assignments */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 43: Annotation Notes */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
		}
		else
		{
			chi = strchr(chs, ',');
		}
		if(chi != NULL)
		{
			printf("Error: Affy_CSVANNOT_To_Reduced_200408, csv annotation loading error!\n");
			exit(EXIT_FAILURE);
		}
		/* strcpy(strTemp, chs); */


		/* write */
		if(vAlign == NULL)
		{
			fprintf(fpOut, "%d\t%s\t-1\t-1\t-1\t?\t-1.0\t%d\t%s\n", 
						nLineId, strProbe, nLocusID, strGenSym);
		}
		else
		{
			pNewAlign = vAlign;
			while(pNewAlign != NULL)
			{
				fprintf(fpOut, "%d\t%s\t%d\t%d\t%d\t%c\t%f\t%d\t%s\n", 
						nLineId, strProbe, pNewAlign->nChr, pNewAlign->nStart, pNewAlign->nEnd-1,
						pNewAlign->chRc, pNewAlign->dIdentity, 
						nLocusID, strGenSym);
				pNewAlign = pNewAlign->pNext;
			}
		}

	
		/* destroy */
		while(vAlign != NULL)
		{
			pNewAlign = vAlign;
			vAlign = vAlign->pNext;
			AffyGenomeAlignDestroy(pNewAlign);
		}
		while(vTrans != NULL)
		{
			pNewAlign = vTrans;
			vTrans = vTrans->pNext;
			AffyGenomeAlignDestroy(pNewAlign);
		}
		for(ni=0; ni<nRefCount; ni++)
		{
			DeleteString(vRefId[ni]);
			vRefId[ni] = NULL;
		}
	}


	/* close file */
	fclose(fpIn);
	fclose(fpOut);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_CSVANNOT_To_Reduced_200408()                                      */
/*  convert affy csv annotation file to reduced annotation                 */
/* ----------------------------------------------------------------------- */ 
int Affy_CSVANNOT_To_Reduced_200408(char strInFile[], char strOutFile[], char strSpecies[])
{
	/* define */
	FILE *fpIn;
	FILE *fpOut;

	char strLine[LONG_LINE_LENGTH];
	char strTemp[MEDLONG_LINE_LENGTH];
	char strChrPos[MEDLONG_LINE_LENGTH];
	char strRefLine[MEDLONG_LINE_LENGTH];
	char *chi,*chs;
	char *chSep,*chp;
	char strProbe[LINE_LENGTH];
	char strGenSym[MEDLONG_LINE_LENGTH];
	char strUniGene[LINE_LENGTH];
	char strLocusID[LINE_LENGTH];
	int nLocusID;

	struct tagAffyGenomeAlign *vAlign;
	struct tagAffyGenomeAlign *vTrans;
	struct tagAffyGenomeAlign *pNewAlign,*pPrevAlign,*pTarAlign;
	struct tagString *vRefId[1024];
	int nRefCount;
	int ni,nLen,nLineId;
	int nAlignWritten;

	
	
	/* init */
	vAlign = NULL;
	vTrans = NULL;
	fpIn = NULL;
	fpOut = NULL;

	/* open file */
	fpIn = fopen(strInFile,"rt");
	if(fpIn == NULL)
	{
		printf("Error: Affy_CSVANNOT_To_Reduced_200408, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = fopen(strOutFile,"wt");
	if(fpOut == NULL)
	{
		fclose(fpIn);
		printf("Error: Affy_CSVANNOT_To_Reduced_200408, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* read */
	nLineId = 0;
	fgets(strLine, LONG_LINE_LENGTH, fpIn);
	while(fgets(strLine, LONG_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		nLineId++;

		/* parse */
		vAlign = NULL;
		vTrans = NULL;

		/* 1: probe */
		if(strchr(strLine, '\"') == strLine)
		{
			chs = strLine+1;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chs = strLine;
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		strcpy(strProbe, chs);
		chs = chi+1;
		printf("%s\n", strProbe); 
		if(strcmp(strProbe, "AFFX-MURINE_B2_at") == 0)
		{
			chs=chs;
		}

		/* debug */
		/* if(strcmp(strProbe, "31807_at") == 0)
		{
			chs = chs;
		} */

		/* 2: GeneChip Array */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 3: Species Scientific Name */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 4: Annotation Date */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 5: Sequence Type */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 6: Sequence Source */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 7: Transcript ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 8: Target Description */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;
		

		/* 9: Representative Public ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 10-a: Archival UniGene Cluster */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		
		/* 10-b: UniGene ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		strcpy(strUniGene, chs);
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 11: Genome Version */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 12: Alignments */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}		
		strcpy(strChrPos, chs);
		chs = chi+1;

		if(strcmp(strChrPos, "---") == 0)
		{	
		}
		else
		{
			chp = strChrPos;
			chSep = strstr(chp, "///");
			while(chSep != NULL)
			{
				*chSep = '\0';
				strcpy(strTemp, chp);

				pNewAlign = NULL;
				pNewAlign = AffyGenomeAlignCreate();
				AffyGenomeAlignInit(pNewAlign, strTemp, strSpecies);
				if(vAlign == NULL)
				{
					vAlign = pNewAlign;
					pPrevAlign = pNewAlign;
				}
				else
				{
					pPrevAlign->pNext = pNewAlign;
					pPrevAlign = pNewAlign;
				}

				chp = chSep+3;
				chSep = strstr(chp, "///");
			}


			strcpy(strTemp, chp);
			pNewAlign = NULL;
			pNewAlign = AffyGenomeAlignCreate();
			AffyGenomeAlignInit(pNewAlign, strTemp, strSpecies);
			if(vAlign == NULL)
			{
				vAlign = pNewAlign;
				pPrevAlign = pNewAlign;
			}
			else
			{
				pPrevAlign->pNext = pNewAlign;
				pPrevAlign = pNewAlign;
			}
		}
			

		/* 13: Overlapping Transcripts */
		/* StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strChrPos, chs);

		chs = chi+1;

		if(strcmp(strChrPos, "---") == 0)
		{	
		}
		else
		{
			chp = strChrPos;
			chSep = strstr(chp, "///");
			while(chSep != NULL)
			{
				*chSep = '\0';
				strcpy(strTemp, chp);

				pNewAlign = NULL;
				pNewAlign = AffyGenomeAlignCreate();
				AffyGenomeTransInit(pNewAlign, strTemp, strSpecies);
				if(vTrans == NULL)
				{
					vTrans = pNewAlign;
					pPrevAlign = pNewAlign;
				}
				else
				{
					pPrevAlign->pNext = pNewAlign;
					pPrevAlign = pNewAlign;
				}

				chp = chSep+3;
				chSep = strstr(chp, "///");
			}


			strcpy(strTemp, chp);
			pNewAlign = NULL;
			pNewAlign = AffyGenomeAlignCreate();
			AffyGenomeTransInit(pNewAlign, strTemp, strSpecies);
			if(vTrans == NULL)
			{
				vTrans = pNewAlign;
				pPrevAlign = pNewAlign;
			}
			else
			{
				pPrevAlign->pNext = pNewAlign;
				pPrevAlign = pNewAlign;
			}
		}
		*/

		/* 14: Gene Title */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 15: Gene Symbol */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strGenSym, chs);
		chs = chi+1;

		if(strcmp(strGenSym, "---") == 0)
			strcpy(strGenSym, "NA");

		/* 16: Chromosomal Location */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 17: UniGene ClusterType */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		/* strcpy(strUniGene, chs); */
		chs = chi+1;

		/* 18: Ensembl */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 19: LocusLink or Entrez Gene */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strLocusID, chs);
		chs = chi+1;

		if(strcmp(strLocusID, "---") == 0)
		{
			nLocusID = -1;
		}
		else if(strstr(strLocusID, "///") != NULL)
		{
			nLocusID = -1;
		}
		else
		{
			nLocusID = atoi(strLocusID);
		}

		/* 20: SwissProt */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 21: EC */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 22: OMIM */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 23: RefSeq Protein ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 24: RefSeq Transcript ID */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chs++;
			chi = strstr(chs, "\",");
			*chi = '\0';
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
			*chi = '\0';
		}
		strcpy(strRefLine, chs);
		chs = chi+1;

		nRefCount = 0;
		ni = 0;
		if(strcmp(strRefLine, "---") == 0)
		{	
		}
		else
		{
			chp = strRefLine;
			chSep = strstr(chp, "///");
			while(chSep != NULL)
			{
				*chSep = '\0';
				strcpy(strTemp, chp);
				StrTrimRight(strTemp);
				StrTrimLeft(strTemp);
				nLen = strlen(strTemp)+1;
				vRefId[ni] = CreateString(nLen);
				strcpy(vRefId[ni]->m_pString, strTemp);
				ni++;

				chp = chSep+3;
				chSep = strstr(chp, "///");
			}

			strcpy(strTemp, chp);
			StrTrimRight(strTemp);
			StrTrimLeft(strTemp);
			nLen = strlen(strTemp)+1;
			vRefId[ni] = CreateString(nLen);
			strcpy(vRefId[ni]->m_pString, strTemp);
			ni++;

			nRefCount = ni;
		}

		
		/* 25: FlyBase */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 26: AGI */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 27: WormBase */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 28: MGI Name */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 29: RGD Name */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 30: SGD accession number */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 31: Gene Ontology Biological Process */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 32: Gene Ontology Cellular Component */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 33: Gene Ontology Molecular Function */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 34: Pathway */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 35: Protein Families */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 36: Protein Domains */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 37: InterPro */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 38: Trans Membrane */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 39: QTL */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 40: Annotation Description */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;
		
		/* 41: Annotation Transcript Cluster */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 42: Transcript Assignments */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
			chi++;
		}
		else
		{
			chi = strchr(chs, ',');
		}
		*chi = '\0';
		/* strcpy(strTemp, chs); */
		chs = chi+1;

		/* 43: Annotation Notes */
		StrTrimLeft(chs);
		if(strchr(chs, '\"') == chs)
		{
			chi = strstr((chs+1), "\",");
		}
		else
		{
			chi = strchr(chs, ',');
		}
		if(chi != NULL)
		{
			printf("Error: Affy_CSVANNOT_To_Reduced_200408, csv annotation loading error!\n");
			exit(EXIT_FAILURE);
		}
		/* strcpy(strTemp, chs); */


		/* write */
		if(nRefCount == 0)
		{
			fprintf(fpOut, "%s\t---\t%d\t%s\n", 
					strProbe, nLocusID, strGenSym);
		}
		else
		{
			for(ni=0; ni<nRefCount; ni++)
			{
				fprintf(fpOut, "%s\t%s\t%d\t%s\n", 
					strProbe, vRefId[ni]->m_pString, nLocusID, strGenSym);
			}
		}


		/* if(vAlign == NULL)
		{
			if(nRefCount == 0)
			{
				fprintf(fpOut, "%s\t-1\t-1\t-1\t?\t-1.0\tNA\t%d\t%s\n", 
					strProbe, nLocusID, strGenSym);
			}
			else
			{
				for(ni=0; ni<nRefCount; ni++)
				{
					fprintf(fpOut, "%s\t-1\t-1\t-1\t?\t-1.0\t%s\t%d\t%s\n", 
						strProbe, vRefId[ni]->m_pString, nLocusID, strGenSym);
				}
			}
		}
		else
		{
			pNewAlign = vAlign;
			while(pNewAlign != NULL)
			{
				if(nRefCount == 0)
				{
					fprintf(fpOut, "%s\t%d\t%d\t%d\t%c\t%f\tNA\t%d\t%s\n", 
						strProbe, pNewAlign->nChr, pNewAlign->nStart, pNewAlign->nEnd-1,
						pNewAlign->chRc, pNewAlign->dIdentity,
						nLocusID, strGenSym);
				}
				else
				{
					nAlignWritten = 0;
					for(ni=0; ni<nRefCount; ni++)
					{
						pTarAlign = NULL;
						pPrevAlign = vTrans;
						while(pPrevAlign != NULL)
						{
							if(strcmp(vRefId[ni]->m_pString, pPrevAlign->strRefId) == 0)
							{
								pTarAlign = pPrevAlign;
								break;
							}
							pPrevAlign = pPrevAlign->pNext;
						}
						
						if(pTarAlign != NULL)
						{
							if( (pTarAlign->nChr == pNewAlign->nChr) &&
								(pTarAlign->chRc == pNewAlign->chRc) )
							{
								if((pTarAlign->nStart >= pNewAlign->nStart) &&
									(pTarAlign->nStart <= pNewAlign->nEnd))
								{
									fprintf(fpOut, "%s\t%d\t%d\t%d\t%c\t%f\t%s\t%d\t%s\n", 
										strProbe, pNewAlign->nChr, pNewAlign->nStart, pNewAlign->nEnd-1,
										pNewAlign->chRc, pNewAlign->dIdentity, vRefId[ni]->m_pString,
										nLocusID, strGenSym);
									nAlignWritten = 1;
								}
								else if((pNewAlign->nStart >= pTarAlign->nStart) &&
									(pNewAlign->nStart <= pTarAlign->nEnd))
								{
									fprintf(fpOut, "%s\t%d\t%d\t%d\t%c\t%f\t%s\t%d\t%s\n", 
										strProbe, pNewAlign->nChr, pNewAlign->nStart, pNewAlign->nEnd-1,
										pNewAlign->chRc, pNewAlign->dIdentity, vRefId[ni]->m_pString,
										nLocusID, strGenSym);
									nAlignWritten = 1;
								}
							}
						}
						else
						{
							fprintf(fpOut, "%s\t%d\t%d\t%d\t%c\t%f\t%s\t%d\t%s\n", 
								strProbe, pNewAlign->nChr, pNewAlign->nStart, pNewAlign->nEnd-1,
								pNewAlign->chRc, pNewAlign->dIdentity, vRefId[ni]->m_pString,
								nLocusID, strGenSym);
							nAlignWritten = 1;
						}						
					}

					if(nAlignWritten == 0)
					{
						fprintf(fpOut, "%s\t%d\t%d\t%d\t%c\t%f\tNA\t%d\t%s\n", 
							strProbe, pNewAlign->nChr, pNewAlign->nStart, pNewAlign->nEnd-1,
							pNewAlign->chRc, pNewAlign->dIdentity,
							nLocusID, strGenSym);
					}
				}

				

				pNewAlign = pNewAlign->pNext;
			}
		} 
		*/
			
		/* destroy */
		while(vAlign != NULL)
		{
			pNewAlign = vAlign;
			vAlign = vAlign->pNext;
			AffyGenomeAlignDestroy(pNewAlign);
		}
		while(vTrans != NULL)
		{
			pNewAlign = vTrans;
			vTrans = vTrans->pNext;
			AffyGenomeAlignDestroy(pNewAlign);
		}
		for(ni=0; ni<nRefCount; ni++)
		{
			DeleteString(vRefId[ni]);
			vRefId[ni] = NULL;
		}
	}


	/* close file */
	fclose(fpIn);
	fclose(fpOut);

	/* return */
	return PROC_FAILURE;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeAlignCreate()                                                */
/*  create genome alignment element                                        */
/* ----------------------------------------------------------------------- */ 
struct tagAffyGenomeAlign *AffyGenomeAlignCreate()
{
	/* define */
	struct tagAffyGenomeAlign *pGA;

	/* create */
	pGA = NULL;
	pGA = (struct tagAffyGenomeAlign *)calloc(1, sizeof(struct tagAffyGenomeAlign));
	if(pGA == NULL)
	{
		printf("Error: cannot create affygenomealign element!\n");
		return NULL;
	}

	/* return */
	return pGA;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeAlignDestroy()                                               */
/*  create genome alignment element                                        */
/* ----------------------------------------------------------------------- */ 
int AffyGenomeAlignDestroy(struct tagAffyGenomeAlign *pGA)
{
	if(pGA == NULL)
		return PROC_SUCCESS;

	pGA->pNext = NULL;
	free(pGA);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeAlignInit()                                                  */
/*  init alignment element                                                 */
/* ----------------------------------------------------------------------- */ 
int AffyGenomeAlignInit(struct tagAffyGenomeAlign *pGA, char strParam[], char strSpecies[])
{
	/* define */
	char *p1,*p2;
	char strLine[MED_LINE_LENGTH];
	char strTemp[LINE_LENGTH];

	/* init */
	if(pGA == NULL)
		return PROC_FAILURE;

	strcpy(strLine, strParam);

	p1 = strstr(strLine, "//");
	*p1 = '\0';
	p1 = p1+2;
	p2 = strstr(p1, "//");
	*p2 = '\0';
	p2 = p2+2;

	strcpy(strTemp, p1);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	pGA->dIdentity = atof(strTemp);
	
	strcpy(strTemp, p2);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	strcpy(pGA->strPhyMap, strTemp);

	StrTrimLeft(strLine);
	StrTrimRight(strLine);

	p1 = strstr(strLine, "(");
	*p1 = '\0';
	p1++;
	p2 = strstr(p1, ")");
	*p2 = '\0';
	strcpy(strTemp, p1);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	if(strcmp(strTemp, "+") == 0)
	{
		pGA->chRc = '+';
	}
	else if(strcmp(strTemp, "-") == 0)
	{
		pGA->chRc = '-';
	}
	else
	{
		pGA->chRc = '?';
	}


	StrTrimRight(strLine);
	p1 = strstr(strLine, ":");
	*p1 = '\0';
	p1++;
	strcpy(strTemp, strLine);
	StrTrimRight(strTemp);
	strcpy(pGA->strChr, strTemp);
	pGA->nChr = Genome_ChromosomeName_To_Index(strTemp, strSpecies);
	p2 = strstr(p1, "-");
	*p2 = '\0';
	p2++;
	strcpy(strTemp, p1);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	pGA->nStart = atoi(strTemp);
	strcpy(strTemp, p2);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	pGA->nEnd = atoi(strTemp);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeTransInit()                                                  */
/*  init alignment element                                                 */
/* ----------------------------------------------------------------------- */ 
int AffyGenomeTransInit(struct tagAffyGenomeAlign *pGA, char strParam[], char strSpecies[])
{
	/* define */
	char *p1,*p2;
	char strLine[MED_LINE_LENGTH];
	char strTemp[LINE_LENGTH];

	/* init */
	if(pGA == NULL)
		return PROC_FAILURE;

	strcpy(strLine, strParam);

	p1 = strstr(strLine, "//");
	*p1 = '\0';
	p1 = p1+2;
	strcpy(strTemp, strLine);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	strcpy(pGA->strRefId, strTemp);


	p2 = strstr(p1, "//");
	*p2 = '\0';
	p2 = p2+2;
	strcpy(strLine, p2);

	StrTrimLeft(strLine);
	StrTrimRight(strLine);

	p1 = strstr(strLine, "(");
	*p1 = '\0';
	p1++;
	p2 = strstr(p1, ")");
	*p2 = '\0';
	strcpy(strTemp, p1);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	if(strcmp(strTemp, "+") == 0)
	{
		pGA->chRc = '+';
	}
	else if(strcmp(strTemp, "-") == 0)
	{
		pGA->chRc = '-';
	}
	else
	{
		pGA->chRc = '?';
	}


	StrTrimRight(strLine);
	p1 = strstr(strLine, ":");
	*p1 = '\0';
	p1++;
	strcpy(strTemp, strLine);
	StrTrimRight(strTemp);
	strcpy(pGA->strChr, strTemp);
	pGA->nChr = Genome_ChromosomeName_To_Index(strTemp, strSpecies);
	p2 = strstr(p1, "-");
	*p2 = '\0';
	p2++;
	strcpy(strTemp, p1);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	pGA->nStart = atoi(strTemp);
	strcpy(strTemp, p2);
	StrTrimLeft(strTemp);
	StrTrimRight(strTemp);
	pGA->nEnd = atoi(strTemp);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_PickRandomControlFromCSV()                                        */
/*  pick random control probesets from affymetrix's CSV annotation file.   */
/* ----------------------------------------------------------------------- */ 
int Affy_PickRandomControlFromCSV(char strInFile[], char strOutFile[], int nControlNum)
{
	/* define */
	FILE *fpIn;
	FILE *fpOut;

	char strLine[LONG_LINE_LENGTH];
	int nProbeNum;
	int ni,nj;
	struct tagString **vProbeName;
	char strProbe[LINE_LENGTH];
	char *chi;

	struct DOUBLEMATRIX *pRandMat;
	struct DOUBLEMATRIX *pSortMat;
	struct LONGMATRIX *pSortId;
	
	/* init */
	fpIn = NULL;
	fpOut = NULL;
	vProbeName = NULL;
	nProbeNum = 0;

	/* open file */
	fpIn = fopen(strInFile,"rt");
	if(fpIn == NULL)
	{
		printf("Error: Affy_PickRandomControlFromCSV, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}
	/* get probenum */
	fgets(strLine, LONG_LINE_LENGTH, fpIn);
	while(fgets(strLine, LONG_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		/* parse */
		nProbeNum++;
	}

	fclose(fpIn);

	/* init probe name vector */
	if(nProbeNum <= 0)
		return PROC_SUCCESS;

	if(nControlNum > nProbeNum)
	{
		printf("Error: Affy_PickRandomControlFromCSV, control probeset number should be <= total probeset number!\n");
		exit(EXIT_FAILURE);
	}

	vProbeName = (struct tagString **)calloc(nProbeNum, sizeof(struct tagString *));
	if(vProbeName == NULL)
	{
		printf("Error: Affy_PickRandomControlFromCSV, cannot allocate memory for loading probeset id's!\n");
		exit(EXIT_FAILURE);
	}

	/* open file again and load probeset id */
	fpIn = NULL;
	fpIn = fopen(strInFile,"rt");
	if(fpIn == NULL)
	{
		printf("Error: Affy_PickRandomControlFromCSV, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}

	/* read probeset id */
	ni = 0;
	fgets(strLine, LONG_LINE_LENGTH, fpIn);
	while(fgets(strLine, LONG_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		/* parse */
		chi = strchr(strLine, ',');
		*chi = '\0';
		strcpy(strProbe, strLine);
		StrTrimRight(strProbe);

		StringAddTail((vProbeName+ni), strProbe);

		/* get next index */
		ni++;
	}

	/* close */
	fclose(fpIn);
	
	if(ni != nProbeNum)
	{
		printf("Error: Affy_PickRandomControlFromCSV, probeset number not match!\n");
		exit(EXIT_FAILURE);
	}

	/* select probesets randomly */
	pRandMat = NULL;
	pSortMat = NULL;
	pSortId = NULL;
	pRandMat = DMRANDU(1, nProbeNum);
	DMSORTMERGEA_0(pRandMat, &pSortMat, &pSortId);

	/* write */
	fpOut = fopen(strOutFile,"wt");
	if(fpOut == NULL)
	{
		printf("Error: Affy_PickRandomControlFromCSV, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	fprintf(fpOut, "probeset_id\n");
	for(ni=0; ni<nControlNum; ni++)
	{
		nj = (int)(LMGETAT(pSortId, 0, ni));
		fprintf(fpOut, "%s\n", vProbeName[nj]->m_pString);
	}

	fclose(fpOut);


	/* release probeset ids */
	DestroyDoubleMatrix(pRandMat);
	DestroyDoubleMatrix(pSortMat);
	DestroyLongMatrix(pSortId);

	for(ni=0; ni<nProbeNum; ni++)
	{
		DeleteString(vProbeName[ni]);
		vProbeName[ni] = NULL;
	}
	free(vProbeName);

	/* return */
	return PROC_SUCCESS;
}


/* ----------------------------------------------------------------------- */ 
/*  Affy_LinkProbesetToAnnotation()                                        */
/*  Link probeset with annotation.                                         */
/* ----------------------------------------------------------------------- */ 
int Affy_LinkProbesetToAnnotation(char strProbesetFile[], char strAnnotationFile[], 
								  char strOutputFile[], int nClassId)
{
	/* define */
	FILE *fpIn = NULL;
	FILE *fpMap = NULL;
	FILE *fpOut = NULL;

	char strLine[MED_LINE_LENGTH];
	char strLine2[MED_LINE_LENGTH];
	char strProbe[LINE_LENGTH];
	char strProbe2[LINE_LENGTH];
	int nLen;
	char *chStart;

	int nLineId, nRank;
	double dScore, dFDR;

	/* open input list */
	fpIn = fopen(strProbesetFile, "rt");
	if(fpIn == NULL)
	{
		printf("Error: Affy_LinkProbesetToAnnotation, cannot open probeset file!\n");
		exit(EXIT_FAILURE);
	}

	fpOut = fopen(strOutputFile, "wt");
	if(fpOut == NULL)
	{
		fclose(fpIn);
		printf("Error: Affy_LinkProbesetToAnnotation, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* load probeset one by one */
	fgets(strLine, MED_LINE_LENGTH, fpIn);
	while(fgets(strLine, MED_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		sscanf(strLine, "%s %d %d %lf %lf ", strProbe, &nLineId, &nRank, &dScore, &dFDR);
		
		fpMap = NULL;
		fpMap = fopen(strAnnotationFile, "rt");
		if(fpMap == NULL)
		{
			printf("Warning: Affy_LinkProbesetToAnnotation, cannot open annotation file!\n");
			continue;
		}

		while(fgets(strLine2, MED_LINE_LENGTH, fpMap) != NULL)
		{
			StrTrimLeft(strLine2);
			StrTrimRight(strLine2);
			if(strLine2[0] == '\0')
				continue;

			sscanf(strLine2, "%s ", strProbe2);
			if(strcmp(strProbe, strProbe2) == 0)
			{
				nLen = strlen(strProbe2);
				chStart = strLine2+nLen;
				StrTrimLeft(chStart);

				fprintf(fpOut, "%s\t%d\t%9.7e\t%9.7e\t%s\n", 
					strProbe, nClassId, dScore, dFDR, chStart);
			}
		}
		fclose(fpMap);
	}

	/* close output file */
	fclose(fpIn);
	fclose(fpOut);

	/* exit */
	return PROC_SUCCESS;
}


/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBar_Intensity()                                               */
/*  load intensity data from *.bar file                                    */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBar_Intensity(char strBarFile[], char strOutPath[])
{
	/* define */
	FILE *fpIn;
	FILE *fpOut;
	struct INTMATRIX *pPos;
	struct DOUBLEMATRIX *pPM;
	struct DOUBLEMATRIX *pMM;
	char strOutFile[LINE_LENGTH];


	char strFiletype[9];
	float fVersion;
	int nSeq;
	int nCol;
	int *vFieldTypes;
	int nTagValPairNum;
	int nTagLen;
	char *vTagName;
	int nTagNameLen;
	char *vTagValPairName;
	int nSeqLen;
	char *vSeqName;
	int nSeqVerLen;
	char *vSeqVersion;
	int nDataPoint;

	int ni,nj,nk;

	float fTemp;
	int nTemp;

	/* load */
	fpIn = NULL;
	fpIn = fopen(strBarFile, "rb");
	if(fpIn == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}

	/* load head */
	fread( strFiletype, sizeof(char), 8, fpIn );
	strFiletype[8] = '\0';
	AFFYBAR_READ_FLOAT(fpIn, &fVersion);
	AFFYBAR_READ_INT(fpIn, &nSeq);
	AFFYBAR_READ_INT(fpIn, &nCol);

		
	/* load field types */
	vFieldTypes = NULL;
	vFieldTypes = (int *)calloc(nCol, sizeof(int));
	if(vFieldTypes == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity, cannot create memory for loading!\n");
		exit(EXIT_FAILURE);
	}
	for(ni=0; ni<nCol; ni++)
	{
		AFFYBAR_READ_INT(fpIn, (vFieldTypes+ni));
	}

	/* load tag-value pair */
	AFFYBAR_READ_INT(fpIn, &nTagValPairNum);
	/* for(ni = 0; ni<nTagValPairNum; ni++) */
	for(ni = 0; ni<1; ni++)
	{
		AFFYBAR_READ_INT(fpIn, &nTagLen);
		vTagName = NULL;
		vTagName = (char *)calloc((nTagLen+1), sizeof(char));
		fread( vTagName, sizeof(char), nTagLen, fpIn);
		vTagName[nTagLen] = '\0';
		free(vTagName);
		AFFYBAR_READ_INT(fpIn, &nTagNameLen);
		vTagValPairName = NULL;
		vTagValPairName = (char *)calloc((nTagNameLen+1), sizeof(char));
		fread( vTagValPairName, sizeof(char), nTagNameLen, fpIn);
		vTagValPairName[nTagNameLen] = '\0';
		free(vTagValPairName);
	}

	for(ni=0; ni<nSeq; ni++)
	{
		AFFYBAR_READ_INT(fpIn, &nSeqLen);
		vSeqName = NULL;
		vSeqName = (char *)calloc((nSeqLen+1), sizeof(char));
		fread( vSeqName, sizeof(char), nSeqLen, fpIn);
		vSeqName[nSeqLen] = '\0'; 
		

		AFFYBAR_READ_INT(fpIn, &nSeqVerLen);
		vSeqVersion = NULL;
		vSeqVersion = (char *)calloc((nSeqVerLen+1), sizeof(char));
		fread( vSeqVersion, sizeof(char), nSeqVerLen, fpIn);
		vSeqVersion[nSeqVerLen] ='\0';
		
		AFFYBAR_READ_INT(fpIn, &nDataPoint);

		/* prepare storage space */
		pPos = NULL;
		pPos = CreateIntMatrix(1, nDataPoint);
		if(pPos == NULL)
		{
			printf("Error: Affy_LoadBar_Intensity, cannot create memory for storing position!\n");
			exit(EXIT_FAILURE);
		}

		pPM = NULL;
		pPM = CreateDoubleMatrix(1, nDataPoint);
		if(pPM == NULL)
		{
			printf("Error: Affy_LoadBar_Intensity, cannot create memory for storing intensity values!\n");
			exit(EXIT_FAILURE);
		}

		pMM = NULL;
		pMM = CreateDoubleMatrix(1, nDataPoint);
		if(pMM == NULL)
		{
			printf("Error: Affy_LoadBar_Intensity, cannot create memory for storing intensity values!\n");
			exit(EXIT_FAILURE);
		}

		/* read data point */
		for(nj=0; nj<nDataPoint; nj++)
		{
			for(nk=0; nk<nCol; nk++)
			{
				switch(vFieldTypes[nk])
				{
					case 1: AFFYBAR_READ_FLOAT(fpIn, &fTemp);
						break;
					case 2: AFFYBAR_READ_INT(fpIn, &nTemp);
						break;
					default: printf("Error: Affy_LoadBar_Intensity, loading error!\n");
						exit(EXIT_FAILURE);
				}

				switch(nk)
				{
					case 0: pPos->pMatElement[nj] = nTemp;
						break;
					case 1: pPM->pMatElement[nj] = fTemp;
						break;
					case 2: pMM->pMatElement[nj] = fTemp;
						break;
					default: printf("Error: Affy_LoadBar_Intensity, loading error!\n");
						exit(EXIT_FAILURE);
				}
			}
		}

		/* save */
		sprintf(strOutFile, "%s%s_intensity.txt", strOutPath, vSeqName);
		fpOut = NULL;
		fpOut = fopen(strOutFile, "wt");
		if(fpOut == NULL)
		{
			printf("Error: Affy_LoadBar_Intensity, cannot open output files!\n");
			exit(EXIT_FAILURE);
		}
		for(nj=0; nj<nDataPoint; nj++)
		{
			fprintf(fpOut, "%d\t%f\t%f\n", pPos->pMatElement[nj],
				pPM->pMatElement[nj], pMM->pMatElement[nj]);
		}

		fclose(fpOut);

		/* destroy */
		free(vSeqName);
		free(vSeqVersion);

		DestroyIntMatrix(pPos);
		DestroyDoubleMatrix(pPM);
		DestroyDoubleMatrix(pMM);
	}

	while(feof(fpIn) == 0)
	{
		fread(strFiletype, sizeof(char), 1, fpIn);
		printf("%c", strFiletype[0]);
	}

	/* close file */
	fclose(fpIn);

	/* destroy */
	free(vFieldTypes);

	/* return */
	return PROC_SUCCESS;
}


/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBar_Intensity_Group()                                         */
/*  load a group of intensity data from *.bar file                         */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBar_Intensity_Group(char strDataPath[], char strBarFileList[], char strOutPath[])
{
	/* define */
	FILE *fpList;
	FILE *fpOut;

	FILE **vfpIn;
	struct tagString **vCondition;
	int nArrayNum;

	char strLine[LINE_LENGTH];
	char strAlias[LINE_LENGTH];
	char strFilePath[LINE_LENGTH];
	
	struct INTMATRIX *pPos;
	struct DOUBLEMATRIX *pPM;
	struct DOUBLEMATRIX *pMM;

	char strFiletype[9];
	float fVersion;
	int nSeq;
	int nCol;
	char strFiletype0[9];
	float fVersion0;
	int nSeq0;
	int nCol0;


	int *vFieldTypes;
	int nTagValPairNum;
	int nTagLen;
	char *vTagName;
	int nTagNameLen;
	char *vTagValPairName;
	int nSeqLen;
	char *vSeqName;
	int nSeqVerLen;
	char *vSeqVersion;

	struct tagString *pSeqName;
	struct tagString *pSeqVersion;

	int nDataPoint, nDataPoint0;

	int ni,nj,nk,na;

	float fTemp;
	double fdiff;
	int nTemp;

	/* load file list */
	sprintf(strLine, "%s%s", strDataPath, strBarFileList);
	fpList = NULL;
	fpList = fopen(strLine, "r");
	if(fpList == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}

	fgets(strLine, LINE_LENGTH, fpList);
	StrTrimLeft(strLine);
	StrTrimRight(strLine);
	nArrayNum = atoi(strLine);
	if(nArrayNum <= 0)
	{
		fclose(fpList);
		printf("Affy_LoadBar_Intensity_Group, no arrays were loaded!\n");
		return PROC_SUCCESS;
	}

	vCondition = NULL;
	vCondition = (struct tagString **)calloc(nArrayNum, sizeof(struct tagString*));
	if(vCondition == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}

	vfpIn = NULL;
	vfpIn = (FILE **)calloc(nArrayNum, sizeof(FILE *));
	if(vfpIn == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}

	ni = 0;
	while(fgets(strLine, LINE_LENGTH, fpList) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		if(ni >= nArrayNum)
		{
			printf("Error: Affy_LoadBar_Intensity_Group, array number not match!\n");
			exit(EXIT_FAILURE);
		}

		sscanf(strLine, "%s %s", strFilePath, strAlias);
		StringAddTail((vCondition+ni), strAlias);
		sprintf(strLine, "%s%s", strDataPath, strFilePath);
		vfpIn[ni] = fopen(strLine, "rb");
		if(vfpIn[ni] == NULL)
		{
			printf("Error: Affy_LoadBar_Intensity_Group, cannot open input file!\n");
			exit(EXIT_FAILURE);
		}

		ni++;
	}

	if(ni != nArrayNum)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, array number not match!\n");
		exit(EXIT_FAILURE);
	}

	/* close list file */
	fclose(fpList);

	/* load head */
	fread( strFiletype, sizeof(char), 8, vfpIn[0] );
	strFiletype[8] = '\0';
	AFFYBAR_READ_FLOAT(vfpIn[0], &fVersion);
	AFFYBAR_READ_INT(vfpIn[0], &nSeq);
	AFFYBAR_READ_INT(vfpIn[0], &nCol);

	for(ni=1; ni<nArrayNum; ni++)
	{
		fread( strFiletype0, sizeof(char), 8, vfpIn[ni] );
		strFiletype0[8] = '\0';
		AFFYBAR_READ_FLOAT(vfpIn[ni], &fVersion0);
		AFFYBAR_READ_INT(vfpIn[ni], &nSeq0);
		AFFYBAR_READ_INT(vfpIn[ni], &nCol0);

		if( (strcmp(strFiletype, strFiletype0) != 0) || (fabs(fVersion0-fVersion)>1e-6)
			|| (nSeq != nSeq0) || (nCol != nCol0) )
		{
			printf("Error: Affy_LoadBar_Intensity_Group, *.bar file not match!\n");
			exit(EXIT_FAILURE);
		}
	}
		
	/* load field types */
	vFieldTypes = NULL;
	vFieldTypes = (int *)calloc(nCol, sizeof(int));
	if(vFieldTypes == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot create memory for loading!\n");
		exit(EXIT_FAILURE);
	}
	for(ni=0; ni<nArrayNum; ni++)
	{
		for(nj=0; nj<nCol; nj++)
		{
			AFFYBAR_READ_INT(vfpIn[ni], (vFieldTypes+nj));
		}
	}

	/* load tag-value pair */
	for(ni=0; ni<nArrayNum; ni++)
	{
		AFFYBAR_READ_INT(vfpIn[ni], &nTagValPairNum);
		/* for(ni = 0; ni<nTagValPairNum; ni++) */
		for(nj=0; nj<1; nj++)
		{
			AFFYBAR_READ_INT(vfpIn[ni], &nTagLen);
			vTagName = NULL;
			vTagName = (char *)calloc((nTagLen+1), sizeof(char));
			fread( vTagName, sizeof(char), nTagLen, vfpIn[ni]);
			vTagName[nTagLen] = '\0';
			free(vTagName);
			AFFYBAR_READ_INT(vfpIn[ni], &nTagNameLen);
			vTagValPairName = NULL;
			vTagValPairName = (char *)calloc((nTagNameLen+1), sizeof(char));
			fread( vTagValPairName, sizeof(char), nTagNameLen, vfpIn[ni]);
			vTagValPairName[nTagNameLen] = '\0';
			free(vTagValPairName);
		}
	}


	/* prepare storage space */
	pPos = NULL;
	pPos = CreateIntMatrix(1, nArrayNum);
	if(pPos == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot create memory for storing position!\n");
		exit(EXIT_FAILURE);
	}

	pPM = NULL;
	pPM = CreateDoubleMatrix(1, nArrayNum);
	if(pPM == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot create memory for storing intensity values!\n");
		exit(EXIT_FAILURE);
	}

	pMM = NULL;
	pMM = CreateDoubleMatrix(1, nArrayNum);
	if(pMM == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot create memory for storing intensity values!\n");
		exit(EXIT_FAILURE);
	}

	fpOut = NULL;
	fpOut = fopen(strOutPath, "wt");
	if(fpOut == NULL)
	{
		printf("Error: Affy_LoadBar_Intensity_Group, cannot open output files!\n");
		exit(EXIT_FAILURE);
	}

	fprintf(fpOut, "chromosome\tposition");
	for(ni=0; ni<nArrayNum; ni++)
	{
		fprintf(fpOut, "\t%s", vCondition[ni]->m_pString);
	}
	fprintf(fpOut, "\n");
	
	
	/* load value */
	for(ni=0; ni<nSeq; ni++)
	{
		nDataPoint = -1;
		pSeqName = NULL;
		pSeqVersion = NULL;

		for(na=0; na<nArrayNum; na++)
		{
			AFFYBAR_READ_INT(vfpIn[na], &nSeqLen);
			vSeqName = NULL;
			vSeqName = (char *)calloc((nSeqLen+1), sizeof(char));
			fread( vSeqName, sizeof(char), nSeqLen, vfpIn[na]);
			vSeqName[nSeqLen] = '\0'; 

			if(pSeqName == NULL)
				StringAddTail(&pSeqName, vSeqName);
			else
			{
				if(strcmp(pSeqName->m_pString, vSeqName) != 0)
				{
					printf("Error: Affy_LoadBar_Intensity_Group, *.bar file sequence name not match!\n");
					exit(EXIT_FAILURE);
				}
			}
			free(vSeqName);
		
			
			AFFYBAR_READ_INT(vfpIn[na], &nSeqVerLen);
			vSeqVersion = NULL;
			vSeqVersion = (char *)calloc((nSeqVerLen+1), sizeof(char));
			fread( vSeqVersion, sizeof(char), nSeqVerLen, vfpIn[na]);
			vSeqVersion[nSeqVerLen] ='\0';

			if(pSeqVersion == NULL)
				StringAddTail(&pSeqVersion, vSeqVersion);
			else
			{
				if(strcmp(pSeqVersion->m_pString, vSeqVersion) != 0)
				{
					printf("Error: Affy_LoadBar_Intensity_Group, *.bar file sequence version not match!\n");
					exit(EXIT_FAILURE);
				}
			}
			free(vSeqVersion);
			
			AFFYBAR_READ_INT(vfpIn[na], &nDataPoint0);
			if(nDataPoint < 0)
			{
				nDataPoint = nDataPoint0;
			}
			else
			{
				if(nDataPoint != nDataPoint0)
				{
					printf("Error: Affy_LoadBar_Intensity_Group, *.bar file number of data point not match!\n");
					exit(EXIT_FAILURE);
				}
			}
		}

		/* read data point */
		for(nj=0; nj<nDataPoint; nj++)
		{
			for(na=0; na<nArrayNum; na++)
			{
				for(nk=0; nk<nCol; nk++)
				{
					switch(vFieldTypes[nk])
					{
						case 1: AFFYBAR_READ_FLOAT(vfpIn[na], &fTemp);
							break;
						case 2: AFFYBAR_READ_INT(vfpIn[na], &nTemp);
							break;
						default: printf("Error: Affy_LoadBar_Intensity_Group, loading error!\n");
							exit(EXIT_FAILURE);
					}

					switch(nk)
					{
						case 0: pPos->pMatElement[na] = nTemp;
							break;
						case 1: pPM->pMatElement[na] = fTemp;
							break;
						case 2: pMM->pMatElement[na] = fTemp;
							break;
						default: printf("Error: Affy_LoadBar_Intensity_Group, loading error!\n");
							exit(EXIT_FAILURE);
					}
				}

				if(pPos->pMatElement[na] != pPos->pMatElement[0])
				{
					printf("Error: Affy_LoadBar_Intensity_Group, chromosomal location not match!\n");
					exit(EXIT_FAILURE);
				}
			}

			/* save */
			fprintf(fpOut, "%s\t%d", pSeqName->m_pString, pPos->pMatElement[0]);
			/* fprintf(fpOut, "%d", pPos->pMatElement[0]); */
			for(na=0; na<nArrayNum; na++)
			{
				fdiff = pPM->pMatElement[na]-pMM->pMatElement[na];
				if(fdiff < 1.0)
					fdiff = 1.0;
				fprintf(fpOut, "\t%9.7e", fdiff);
				/* fprintf(fpOut, "\t%9.7e", pPM->pMatElement[na]); */
			}
			fprintf(fpOut, "\n");

		}
		
		/* destroy */
		DeleteString(pSeqName);
		DeleteString(pSeqVersion);
	}

	fclose(fpOut);

	for(ni=0; ni<nArrayNum; ni++)
	{
		printf("file %d (%s) ", ni+1, vCondition[ni]->m_pString);
		while(feof(vfpIn[ni]) == 0)
		{
			fread(strFiletype, sizeof(char), 1, vfpIn[ni]);
			printf("%c", strFiletype[0]);
		}
	}
	
	DestroyIntMatrix(pPos);
	DestroyDoubleMatrix(pPM);
	DestroyDoubleMatrix(pMM);


	/* close file */
	for(ni=0; ni<nArrayNum; ni++)
	{
		fclose(vfpIn[ni]);
		DeleteString(vCondition[ni]);
		vCondition[ni] = NULL;
	}
	free(vfpIn);
	free(vCondition);

	/* destroy */
	free(vFieldTypes);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBPMAP()                                                       */
/*  load bpmap data from *.bar file.                                       */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBPMAP(char strInFile[], char strOutFile[])
{
	/* define */
	FILE *fpIn;
	FILE *fpOut;

	/* variables */
	char strFileType[9];
	float fVersion;
	unsigned long nSeqNum;

	/* seq info */
	struct tagString **vSeqName;
	struct INTMATRIX *vProbePairNum;
	unsigned long nSeqNameLen;
	unsigned long nProbePairNum;

	/* map info */
	struct INTMATRIX *vSeqID;
	unsigned long nSeqID;
	int nProbeNum;
	struct tagAffyBpMapUnit *pNewUnit,*pCUnit,*pPUnit;
	struct tagAffyBpMapUnit *pUnitList;

	/* count */
	int ni,nj;
	int nEndPos;
	int nIgnore;

	/* load */
	fpIn = NULL;
	fpIn = fopen(strInFile, "rb");
	if(fpIn == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = NULL;
	fpOut = fopen(strOutFile, "wt");
	if(fpOut == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* load head */
	fread( strFileType, sizeof(char), 8, fpIn );
	strFileType[8] = '\0';
	AFFYBAR_READ_FLOAT(fpIn, &fVersion);
	AFFYBAR_READ_ULONG(fpIn, &nSeqNum);

	/* load sequence names */
	vSeqName = NULL;
	vSeqName = (struct tagString **)calloc(nSeqNum, sizeof(struct tagString *));
	if(vSeqName == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot load sequence name!\n");
		exit(EXIT_FAILURE);
	}
	vProbePairNum = NULL;
	vProbePairNum = CreateIntMatrix(nSeqNum,1);
	if(vProbePairNum == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot load probe pair number!\n");
		exit(EXIT_FAILURE);
	}
	
	for(ni=0; ni<(int)nSeqNum; ni++)
	{
		AFFYBAR_READ_ULONG(fpIn, &nSeqNameLen);
		vSeqName[ni] = CreateString(nSeqNameLen);
		fread( vSeqName[ni]->m_pString, sizeof(char), nSeqNameLen, fpIn );
		vSeqName[ni]->m_pString[nSeqNameLen] = '\0';
		AFFYBAR_READ_ULONG(fpIn, &nProbePairNum);
		vProbePairNum->pMatElement[ni] = (int)nProbePairNum;
	}

	/* load map */
	vSeqID = NULL;
	vSeqID = CreateIntMatrix(nSeqNum,1);
	if(vSeqID == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot load sequence id!\n");
		exit(EXIT_FAILURE);
	}
	for(ni=0; ni<(int)nSeqNum; ni++)
	{
		/* load seq id */
		AFFYBAR_READ_ULONG(fpIn, &nSeqID);
		nProbeNum = vProbePairNum->pMatElement[ni];
		pUnitList = NULL;
		
		/* load probes */
		for(nj=0; nj<nProbeNum; nj++)
		{
			/* load new unit */
			pNewUnit = NULL;
			pNewUnit = AffyBpMapUnitCreate();
			AffyBpMapUnitLoad(pNewUnit, fpIn);
			
			/* add to unitlist */
			if(pUnitList == NULL)
			{
				pUnitList = pNewUnit;
			}
			else
			{
				nIgnore = 0;
				pCUnit = pUnitList;
				while(pCUnit != NULL)
				{
					/* if same position */
					if(pNewUnit->nPos == pCUnit->nPos)
					{
						pCUnit->nDepthNum += 1;
						nIgnore = 1;
					}
					pCUnit = pCUnit->pNext;
				}

				if(nIgnore == 1)
				{
					AffyBpMapUnitDestroy(pNewUnit);
				}
				else
				{
					pCUnit = pUnitList;
					pPUnit = NULL;
					while(pCUnit != NULL)
					{
						/* if local repeat */
						if(pNewUnit->nPos-pCUnit->nPos <= LOCAL_REPEAT_RANGE)
						{
							if(strcmp(pCUnit->strProbeSeq, pNewUnit->strProbeSeq) == 0)
							{
								pCUnit->nRepeatNum += 1;
								pNewUnit->nRepeatNum += 1;
							}
						}

						/* get next */
						pPUnit = pCUnit;
						pCUnit = pPUnit->pNext;
					}

					pPUnit->pNext = pNewUnit;
					nEndPos = pNewUnit->nPos;

					/* write old unit */
					while(pUnitList != NULL)
					{
						/* if local repeat */
						if(nEndPos-pUnitList->nPos > LOCAL_REPEAT_RANGE)
						{
							pCUnit = pUnitList;
							pUnitList = pCUnit->pNext;
							pCUnit->pNext = NULL;

							fprintf(fpOut, "%d\t%d\t%d\t%s\n", pCUnit->nPos, pCUnit->nDepthNum, pCUnit->nRepeatNum, pCUnit->strProbeSeq);
							AffyBpMapUnitDestroy(pCUnit);
						}
						else
						{
							break;
						}
					}
				}
			}			
		}

		/* clear unit list */
		while(pUnitList != NULL)
		{
			pCUnit = pUnitList;
			pUnitList = pCUnit->pNext;
			pCUnit->pNext = NULL;

			fprintf(fpOut, "%d\t%d\t%d\t%s\n", pCUnit->nPos, pCUnit->nDepthNum, pCUnit->nRepeatNum, pCUnit->strProbeSeq);
			AffyBpMapUnitDestroy(pCUnit);
		}
	}

	/* load tail if any */
	while(feof(fpIn) == 0)
	{
		fread(strFileType, sizeof(char), 1, fpIn);
		printf("%c", strFileType[0]);
	}

	/* clear memeory */
	for(ni=0; ni<(int)nSeqNum; ni++)
	{
		DeleteString(vSeqName[ni]);
		vSeqName[ni] = NULL;
	}
	free(vSeqName);
	DestroyIntMatrix(vProbePairNum);
	DestroyIntMatrix(vSeqID);

	/* close file */
	fclose(fpIn);
	fclose(fpOut);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBPMAP_TileMap()                                               */
/*  load bpmap data from *.bar file, tilemap format                        */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBPMAP_TileMap(char strInFile[], char strPosFile[], 
						   char strMaskFile[], int *pnTotalProbeNum)
{
	/* define */
	FILE *fpIn;
	FILE *fpOut;
	FILE *fpPos;

	/* variables */
	char strFileType[9];
	float fVersion;
	unsigned long nSeqNum;

	/* seq info */
	struct tagString **vSeqName;
	struct INTMATRIX *vProbePairNum;
	unsigned long nSeqNameLen;
	unsigned long nProbePairNum;

	/* map info */
	struct INTMATRIX *vSeqID;
	unsigned long nSeqID;
	int nProbeNum;
	struct tagAffyBpMapUnit *pNewUnit,*pCUnit,*pPUnit;
	struct tagAffyBpMapUnit *pUnitList;

	/* count */
	int ni,nj;
	int nEndPos;
	int nIgnore;

	/* load */
	fpIn = NULL;
	fpIn = fopen(strInFile, "rb");
	if(fpIn == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot open input file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = NULL;
	fpOut = fopen(strMaskFile, "w");
	if(fpOut == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}
	fpPos = NULL;
	fpPos = fopen(strPosFile, "w");
	if(fpPos == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* load head */
	fread( strFileType, sizeof(char), 8, fpIn );
	strFileType[8] = '\0';
	AFFYBAR_READ_FLOAT(fpIn, &fVersion);
	AFFYBAR_READ_ULONG(fpIn, &nSeqNum);

	/* load sequence names */
	vSeqName = NULL;
	vSeqName = (struct tagString **)calloc(nSeqNum, sizeof(struct tagString *));
	if(vSeqName == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot load sequence name!\n");
		exit(EXIT_FAILURE);
	}
	vProbePairNum = NULL;
	vProbePairNum = CreateIntMatrix(nSeqNum,1);
	if(vProbePairNum == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot load probe pair number!\n");
		exit(EXIT_FAILURE);
	}
	
	*pnTotalProbeNum = 0;
	for(ni=0; ni<(int)nSeqNum; ni++)
	{
		AFFYBAR_READ_ULONG(fpIn, &nSeqNameLen);
		vSeqName[ni] = CreateString(nSeqNameLen);
		fread( vSeqName[ni]->m_pString, sizeof(char), nSeqNameLen, fpIn );
		vSeqName[ni]->m_pString[nSeqNameLen] = '\0';
		AFFYBAR_READ_ULONG(fpIn, &nProbePairNum);
		vProbePairNum->pMatElement[ni] = (int)nProbePairNum;
		*pnTotalProbeNum += (int)nProbePairNum;
	}

	/* load map */
	vSeqID = NULL;
	vSeqID = CreateIntMatrix(nSeqNum,1);
	if(vSeqID == NULL)
	{
		printf("Error: Affy_LoadBPMAP, cannot load sequence id!\n");
		exit(EXIT_FAILURE);
	}
	for(ni=0; ni<(int)nSeqNum; ni++)
	{
		/* load seq id */
		AFFYBAR_READ_ULONG(fpIn, &nSeqID);
		nProbeNum = vProbePairNum->pMatElement[ni];
		pUnitList = NULL;
		
		/* load probes */
		for(nj=0; nj<nProbeNum; nj++)
		{
			/* load new unit */
			pNewUnit = NULL;
			pNewUnit = AffyBpMapUnitCreate();
			AffyBpMapUnitLoad(pNewUnit, fpIn);
			fprintf(fpPos, "%s\t%d\n", vSeqName[ni]->m_pString, pNewUnit->nPos);
			
			/* add to unitlist */
			if(pUnitList == NULL)
			{
				pUnitList = pNewUnit;
			}
			else
			{
				nIgnore = 0;
				pCUnit = pUnitList;
				while(pCUnit != NULL)
				{
					/* if same position */
					if(pNewUnit->nPos == pCUnit->nPos)
					{
						pCUnit->nDepthNum += 1;
						nIgnore = 1;
					}
					pCUnit = pCUnit->pNext;
				}

				if(nIgnore == 1)
				{
					AffyBpMapUnitDestroy(pNewUnit);
				}
				else
				{
					pCUnit = pUnitList;
					pPUnit = NULL;
					while(pCUnit != NULL)
					{
						/* if local repeat */
						if(pNewUnit->nPos-pCUnit->nPos <= LOCAL_REPEAT_RANGE)
						{
							if(strcmp(pCUnit->strProbeSeq, pNewUnit->strProbeSeq) == 0)
							{
								pCUnit->nRepeatNum += 1;
								pNewUnit->nRepeatNum += 1;
							}
						}

						/* get next */
						pPUnit = pCUnit;
						pCUnit = pPUnit->pNext;
					}

					pPUnit->pNext = pNewUnit;
					nEndPos = pNewUnit->nPos;

					/* write old unit */
					while(pUnitList != NULL)
					{
						/* if local repeat */
						if(nEndPos-pUnitList->nPos > LOCAL_REPEAT_RANGE)
						{
							pCUnit = pUnitList;
							pUnitList = pCUnit->pNext;
							pCUnit->pNext = NULL;

							fprintf(fpOut, "%s\t%d\t%d\t%d\t%s\n", vSeqName[ni]->m_pString, pCUnit->nPos, pCUnit->nDepthNum, pCUnit->nRepeatNum, pCUnit->strProbeSeq);
							AffyBpMapUnitDestroy(pCUnit);
						}
						else
						{
							break;
						}
					}
				}
			}			
		}

		/* clear unit list */
		while(pUnitList != NULL)
		{
			pCUnit = pUnitList;
			pUnitList = pCUnit->pNext;
			pCUnit->pNext = NULL;

			fprintf(fpOut, "%s\t%d\t%d\t%d\t%s\n", vSeqName[ni]->m_pString, pCUnit->nPos, pCUnit->nDepthNum, pCUnit->nRepeatNum, pCUnit->strProbeSeq);
			AffyBpMapUnitDestroy(pCUnit);
		}
	}

	/* load tail if any */
	while(feof(fpIn) == 0)
	{
		fread(strFileType, sizeof(char), 1, fpIn);
		printf("%c", strFileType[0]);
	}

	/* clear memeory */
	for(ni=0; ni<(int)nSeqNum; ni++)
	{
		DeleteString(vSeqName[ni]);
		vSeqName[ni] = NULL;
	}
	free(vSeqName);
	DestroyIntMatrix(vProbePairNum);
	DestroyIntMatrix(vSeqID);

	/* close file */
	fclose(fpIn);
	fclose(fpOut);
	fclose(fpPos);

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitCreate()                                                  */
/*  create bpmap unit.                                                     */
/* ----------------------------------------------------------------------- */ 
struct tagAffyBpMapUnit *AffyBpMapUnitCreate()
{
	/* define */
	struct tagAffyBpMapUnit *pUnit;
	
	/* create */
	pUnit = NULL;
	pUnit = (struct tagAffyBpMapUnit *)calloc(1, sizeof(struct tagAffyBpMapUnit));
	if(pUnit == NULL)
	{
		printf("Error: AffyBpMapUnitCreate, cannot create bpmap structure!\n");
		exit(EXIT_FAILURE);
	}

	/* return */
	return pUnit;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitDestroy()                                                 */
/*  delete bpmap unit.                                                     */
/* ----------------------------------------------------------------------- */ 
int AffyBpMapUnitDestroy(struct tagAffyBpMapUnit *pUnit)
{
	/* destroy */
	if(pUnit != NULL)
	{
		pUnit->pNext = NULL;
		free(pUnit);
		pUnit = NULL;
	}

	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitLoad()                                                    */
/*  load bpmap unit.                                                       */
/* ----------------------------------------------------------------------- */ 
int AffyBpMapUnitLoad(struct tagAffyBpMapUnit *pUnit, FILE *fpIn)
{
	/* define */
	char strTemp[PROBE_BPMAP_LEN];
	int ni,nj,nk;
	int nProbeLen;
	char bChar;

	/* load */
	if((pUnit == NULL) || (fpIn == NULL))
	{
		printf("Error: cannot read bpmap from file!\n");
		exit(EXIT_FAILURE);
	}

	nProbeLen = TILE_PROBE_LEN-1;
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nPMX));
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nPMY));
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nMMX));
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nMMY));
	fread( &(pUnit->bProbeLen), sizeof(unsigned char), 1, fpIn );
	fread( strTemp, sizeof(char), PROBE_BPMAP_LEN, fpIn );
	nk = 0;
	for(ni=0; ni<PROBE_BPMAP_LEN; ni++)
	{
		for(nj=0; (nj<4 && nk<nProbeLen); nj++)
		{
			bChar = (strTemp[ni] >> (2*(3-nj))) & 0x03;
			switch(bChar)
			{
				case DNA_BASE_A:
				  pUnit->strProbeSeq[nk] = 'A';
				  break;
				case DNA_BASE_C:
				  pUnit->strProbeSeq[nk] = 'C';
				  break;
				case DNA_BASE_G:
				  pUnit->strProbeSeq[nk] = 'G';
				  break;
				case DNA_BASE_T:
				  pUnit->strProbeSeq[nk] = 'T';
				  break;
				default:
				  pUnit->strProbeSeq[nk] = 'N';
				  break;
			}
			nk++;
		}
	}
	pUnit->strProbeSeq[nk] = '\0';
	AFFYBAR_READ_FLOAT(fpIn, &(pUnit->fMatchScore));
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nPos));
	fread( &(pUnit->bStrand), sizeof(unsigned char), 1, fpIn );
	pUnit->nRepeatNum = 1;
	pUnit->nDepthNum = 1;


	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitLoad_v3()                                                 */
/*  load bpmap unit.                                                       */
/* ----------------------------------------------------------------------- */ 
int AffyBpMapUnitLoad_v3(struct tagAffyBpMapUnit *pUnit, FILE *fpIn, int nPMType)
{
	/* define */
	char strTemp[PROBE_BPMAP_LEN];
	int ni,nj,nk;
	int nProbeLen;
	char bChar;

	/* load */
	if((pUnit == NULL) || (fpIn == NULL))
	{
		printf("Error: cannot read bpmap from file!\n");
		exit(EXIT_FAILURE);
	}

	nProbeLen = TILE_PROBE_LEN-1;
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nPMX));
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nPMY));
	if(nPMType == 0)
	{
		AFFYBAR_READ_ULONG(fpIn, &(pUnit->nMMX));
		AFFYBAR_READ_ULONG(fpIn, &(pUnit->nMMY));
	}
	fread( &(pUnit->bProbeLen), sizeof(unsigned char), 1, fpIn );
	fread( strTemp, sizeof(char), PROBE_BPMAP_LEN, fpIn );
	nk = 0;
	for(ni=0; ni<PROBE_BPMAP_LEN; ni++)
	{
		for(nj=0; (nj<4 && nk<nProbeLen); nj++)
		{
			bChar = (strTemp[ni] >> (2*(3-nj))) & 0x03;
			switch(bChar)
			{
				case DNA_BASE_A:
				  pUnit->strProbeSeq[nk] = 'A';
				  break;
				case DNA_BASE_C:
				  pUnit->strProbeSeq[nk] = 'C';
				  break;
				case DNA_BASE_G:
				  pUnit->strProbeSeq[nk] = 'G';
				  break;
				case DNA_BASE_T:
				  pUnit->strProbeSeq[nk] = 'T';
				  break;
				default:
				  pUnit->strProbeSeq[nk] = 'N';
				  break;
			}
			nk++;
		}
	}
	pUnit->strProbeSeq[nk] = '\0';
	AFFYBAR_READ_FLOAT(fpIn, &(pUnit->fMatchScore));
	AFFYBAR_READ_ULONG(fpIn, &(pUnit->nPos));
	fread( &(pUnit->bStrand), sizeof(unsigned char), 1, fpIn );
	pUnit->nRepeatNum = 1;
	pUnit->nDepthNum = 1;


	/* return */
	return PROC_SUCCESS;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_BPMAPFilter_GTrans()                                              */
/*  Filter out repeat and bad probes with bpmap informaiton.               */
/* ----------------------------------------------------------------------- */ 
int Affy_BPMAPFilter_GTrans(char strInFile[], char strRefFile[], char strOutFile[])
{
	/* define */
	FILE *fpIn;
	FILE *fpRef;
	FILE *fpOut;

	/* load line */
	char strLine[LINE_LENGTH];
	int nRefPos,nPos;
	int nDepthNum,nRepeatNum;
	char strProbeSeq[LINE_LENGTH];
	double dValue;
	double dMean,dN;
	int nInFileEnd,nRefFileEnd;
	int nWritePos;
	
	/* open file */
	fpIn = NULL;
	fpIn = fopen(strInFile, "r");
	if(fpIn == NULL)
	{
		printf("Error: Affy_BPMAPFilter_GTrans, cannot read input file!\n");
		exit(EXIT_FAILURE);
	}
	fpRef = NULL;
	fpRef = fopen(strRefFile, "r");
	if(fpRef == NULL)
	{
		printf("Error: Affy_BPMAPFilter_GTrans, cannot read reference file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = NULL;
	fpOut = fopen(strOutFile, "w");
	if(fpOut == NULL)
	{
		printf("Error: Affy_BPMAPFilter_GTrans, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* read files */
	nInFileEnd = 1;
	nRefFileEnd = 1;
	nPos = 0;
	nRefPos = 0;
	dValue = 0.0;
	nDepthNum = 0;
	nRepeatNum = 0;
	
	while(fgets(strLine, LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if((strLine[0] != '#') && (strLine[0] != '\0'))
		{
			nInFileEnd = 0;
			break;
		}
	}
	
	if(nInFileEnd == 0)
	{
		sscanf(strLine, "%d %lf", &nPos, &dValue);
	}

	while(fgets(strLine, LINE_LENGTH, fpRef) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] != '\0')
		{
			nRefFileEnd = 0;
			break;
		}
	}
	
	if(nRefFileEnd == 0)
	{
		sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
	}
	
	dMean = 0.0;
	dN = 0.0;
	while((nInFileEnd == 0) && (nRefFileEnd == 0))
	{
		/* if equal position */
		if(nPos == nRefPos)
		{
			dMean += dValue;
			dN += 1.0;

			nInFileEnd = 1;
			while(fgets(strLine, LINE_LENGTH, fpIn) != NULL)
			{
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				if((strLine[0] != '#') && (strLine[0] != '\0'))
				{
					nInFileEnd = 0;
					break;
				}
			}
			
			if(nInFileEnd == 0)
			{
				sscanf(strLine, "%d %lf", &nPos, &dValue);
			}
		}
		/* if pos > refpos */
		else if(nPos > nRefPos)
		{
			if(dN > 0.0)
			{
				dMean /= dN;
				nWritePos = nRefPos;
				if(nRepeatNum <= 1)
				{
					fprintf(fpOut, "%d\t%lf\n", nRefPos, dMean);
				}
			}

			dMean = 0.0;
			dN = 0.0;

			nRefFileEnd = 1;
			while(fgets(strLine, LINE_LENGTH, fpRef) != NULL)
			{
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				if(strLine[0] != '\0')
				{
					nRefFileEnd = 0;
					break;
				}
			}
			
			if(nRefFileEnd == 0)
			{
				sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
			}
		}
		else
		{
			nInFileEnd = 1;
			while(fgets(strLine, LINE_LENGTH, fpIn) != NULL)
			{
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				if((strLine[0] != '#') && (strLine[0] != '\0'))
				{
					nInFileEnd = 0;
					break;
				}
			}
			
			if(nInFileEnd == 0)
			{
				sscanf(strLine, "%d %lf", &nPos, &dValue);
			}
		}
	}

	if( (nWritePos < nRefPos) && (dN > 0.0))
	{
		dMean /= dN;
		if(nRepeatNum <= 1)
		{
			fprintf(fpOut, "%d\t%lf\n", nRefPos, dMean);
		}
	}

	/* close file */
	fclose(fpIn);
	fclose(fpRef);
	fclose(fpOut);


	/* return */
	return PROC_SUCCESS;
}


/* ----------------------------------------------------------------------- */ 
/*  Affy_BPMAPFilter_ORI()                                                 */
/*  Filter out repeat and bad probes with bpmap informaiton.               */
/* ----------------------------------------------------------------------- */ 
int Affy_BPMAPFilter_ORI(char strInFile[], char strRefFile[], char strOutFile[])
{
	/* define */
	FILE *fpIn;
	FILE *fpRef;
	FILE *fpOut;

	/* load line */
	char strLine[LINE_LENGTH];
	int nRefPos,nPos;
	double dPos;
	int nDepthNum,nRepeatNum;
	char strProbeSeq[LINE_LENGTH];
	double dValue;
	double dMean,dN;
	int nInFileEnd,nRefFileEnd;
	int nWritePos;
	
	/* open file */
	fpIn = NULL;
	fpIn = fopen(strInFile, "r");
	if(fpIn == NULL)
	{
		printf("Error: Affy_BPMAPFilter_GTrans, cannot read input file!\n");
		exit(EXIT_FAILURE);
	}
	fpRef = NULL;
	fpRef = fopen(strRefFile, "r");
	if(fpRef == NULL)
	{
		printf("Error: Affy_BPMAPFilter_GTrans, cannot read reference file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = NULL;
	fpOut = fopen(strOutFile, "w");
	if(fpOut == NULL)
	{
		printf("Error: Affy_BPMAPFilter_GTrans, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* read files */
	nInFileEnd = 1;
	nRefFileEnd = 1;
	nPos = 0;
	nRefPos = 0;
	dValue = 0.0;
	nDepthNum = 0;
	nRepeatNum = 0;
	
	while(fgets(strLine, LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if((strLine[0] != '#') && (strLine[0] != '\0'))
		{
			nInFileEnd = 0;
			break;
		}
	}
	
	if(nInFileEnd == 0)
	{
		sscanf(strLine, "%lf %lf", &dPos, &dValue);
		nPos = (int)dPos;
	}

	while(fgets(strLine, LINE_LENGTH, fpRef) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] != '\0')
		{
			nRefFileEnd = 0;
			break;
		}
	}
	
	if(nRefFileEnd == 0)
	{
		sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
	}
	
	dMean = 0.0;
	dN = 0.0;
	while((nInFileEnd == 0) && (nRefFileEnd == 0))
	{
		/* if equal position */
		if(nPos == nRefPos)
		{
			dMean += dValue;
			dN += 1.0;

			nInFileEnd = 1;
			while(fgets(strLine, LINE_LENGTH, fpIn) != NULL)
			{
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				if((strLine[0] != '#') && (strLine[0] != '\0'))
				{
					nInFileEnd = 0;
					break;
				}
			}
			
			if(nInFileEnd == 0)
			{
				sscanf(strLine, "%lf %lf", &dPos, &dValue);
				nPos = (int)dPos;
			}
		}
		/* if pos > refpos */
		else if(nPos > nRefPos)
		{
			if(dN > 0.0)
			{
				if((int)dN != nDepthNum)
				{
					printf("Warning: Affy_BPMAPFilter_ORI, depth number not consistent!\n");
				}

				dMean /= dN;
				nWritePos = nRefPos;
				if(nRepeatNum <= 1)
				{
					fprintf(fpOut, "%d\t%lf\n", nRefPos, dMean);
				}
			}

			dMean = 0.0;
			dN = 0.0;

			nRefFileEnd = 1;
			while(fgets(strLine, LINE_LENGTH, fpRef) != NULL)
			{
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				if(strLine[0] != '\0')
				{
					nRefFileEnd = 0;
					break;
				}
			}
			
			if(nRefFileEnd == 0)
			{
				sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
			}
		}
		else
		{
			nInFileEnd = 1;
			while(fgets(strLine, LINE_LENGTH, fpIn) != NULL)
			{
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				if((strLine[0] != '#') && (strLine[0] != '\0'))
				{
					nInFileEnd = 0;
					break;
				}
			}
			
			if(nInFileEnd == 0)
			{
				sscanf(strLine, "%lf %lf", &dPos, &dValue);
				nPos = (int)dPos;
			}
		}
	}

	if( (nWritePos < nRefPos) && (dN > 0.0))
	{
		if((int)dN != nDepthNum)
		{
			printf("Warning: Affy_BPMAPFilter_ORI, depth number not consistent!\n");
		}
		dMean /= dN;
		if(nRepeatNum <= 1)
		{
			fprintf(fpOut, "%d\t%lf\n", nRefPos, dMean);
		}
	}

	/* close file */
	fclose(fpIn);
	fclose(fpRef);
	fclose(fpOut);


	/* return */
	return PROC_SUCCESS;
}


/* ----------------------------------------------------------------------- */ 
/*  Affy_BPMAPFilter_DATA()                                                 */
/*  Filter out repeat and bad probes with bpmap informaiton.               */
/* ----------------------------------------------------------------------- */ 
int Affy_BPMAPFilter_DATA(char strInFile[], char strRefFile[], char strOutFile[])
{
	/* define */
	FILE *fpIn;
	FILE *fpRef;
	FILE *fpOut;

	/* load line */
	char strLine[LONG_LINE_LENGTH];
	char strDataLine[LONG_LINE_LENGTH];
	int nRefPos,nPos;
	double dPos;
	int nDepthNum,nRepeatNum;
	char strProbeSeq[LINE_LENGTH];
	double dValue;
	double dMean,dN;
	int nInFileEnd,nRefFileEnd;
	int nWritePos;
	
	/* open file */
	fpIn = NULL;
	fpIn = fopen(strInFile, "r");
	if(fpIn == NULL)
	{
		printf("Error: Affy_BPMAPFilter_DATA, cannot read input file!\n");
		exit(EXIT_FAILURE);
	}
	fpRef = NULL;
	fpRef = fopen(strRefFile, "r");
	if(fpRef == NULL)
	{
		printf("Error: Affy_BPMAPFilter_DATA, cannot read reference file!\n");
		exit(EXIT_FAILURE);
	}
	fpOut = NULL;
	fpOut = fopen(strOutFile, "w");
	if(fpOut == NULL)
	{
		printf("Error: Affy_BPMAPFilter_DATA, cannot open output file!\n");
		exit(EXIT_FAILURE);
	}

	/* read files */
	nInFileEnd = 1;
	nRefFileEnd = 1;
	nPos = 0;
	nRefPos = 0;
	dValue = 0.0;
	nDepthNum = 0;
	nRepeatNum = 0;
	
	while(fgets(strLine, LONG_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if((strLine[0] != '#') && (strLine[0] != '\0'))
		{
			nInFileEnd = 0;
			break;
		}
	}
	
	if(nInFileEnd == 0)
	{
		sscanf(strLine, "%lf", &dPos);
		strcpy(strDataLine, strLine);
		nPos = (int)dPos;
	}

	while(fgets(strLine, LONG_LINE_LENGTH, fpRef) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] != '\0')
		{
			nRefFileEnd = 0;
			break;
		}
	}
	
	if(nRefFileEnd == 0)
	{
		sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
	}
	
	while((nInFileEnd == 0) && (nRefFileEnd == 0))
	{
		if(nRefPos == nPos)
		{
			if(nRepeatNum <= 1)
			{
				fprintf(fpOut, "%s\n", strDataLine);
			}
			if(fgets(strLine, LONG_LINE_LENGTH, fpIn) == NULL)
			{
				nInFileEnd = 1;
				break;
			}
			StrTrimLeft(strLine);
			StrTrimRight(strLine);
			sscanf(strLine, "%lf", &dPos);
			strcpy(strDataLine, strLine);
			nPos = (int)dPos;

			if(fgets(strLine, LONG_LINE_LENGTH, fpRef) == NULL)
			{
				nRefFileEnd = 1;
				break;
			}
			StrTrimLeft(strLine);
			StrTrimRight(strLine);
			sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
		}
		else if(nPos < nRefPos)
		{
			while(nPos < nRefPos)
			{
				if(fgets(strLine, LONG_LINE_LENGTH, fpIn) == NULL)
				{
					nInFileEnd = 1;
					break;
				}
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				sscanf(strLine, "%lf", &dPos);
				strcpy(strDataLine, strLine);
				nPos = (int)dPos;
			}
		}
		else
		{
			while(nRefPos < nPos)
			{
				if(fgets(strLine, LONG_LINE_LENGTH, fpRef) == NULL)
				{
					nRefFileEnd = 1;
					break;
				}
				StrTrimLeft(strLine);
				StrTrimRight(strLine);
				sscanf(strLine, "%d %d %d %s", &nRefPos, &nDepthNum, &nRepeatNum, strProbeSeq);
			}
		}
	}
		

	/* close file */
	fclose(fpIn);
	fclose(fpRef);
	fclose(fpOut);


	/* return */
	return PROC_SUCCESS;
}


/* ----------------------------------------------------------------------- */ 
/*  Functions for loading affy binary files.                               */
/* ----------------------------------------------------------------------- */ 
void AFFYBAR_READ_INT(FILE *fpIn, int *value)
{
	char str[INT_SIZE];
	char str2[INT_SIZE];
	int ni;
	
	fread(str, sizeof(char), INT_SIZE, fpIn);
	if(AFFY_IS_BIG_ENDIAN == 1)
	{
		for(ni=0; ni<INT_SIZE; ni++)
			str2[ni] = str[INT_SIZE-1-ni];
		memcpy(value, str2, INT_SIZE);
	}
	else
	{
		memcpy(value, str, INT_SIZE);
	}
}

void AFFYBAR_READ_ULONG(FILE *fpIn, unsigned long *value)
{
	char str[ULONG_SIZE];
	char str2[ULONG_SIZE];
	int ni;
	
	fread(str, sizeof(char), ULONG_SIZE, fpIn);
	if(AFFY_IS_BIG_ENDIAN == 1)
	{
		for(ni=0; ni<ULONG_SIZE; ni++)
			str2[ni] = str[ULONG_SIZE-1-ni];
		memcpy(value, str2, ULONG_SIZE);
	}
	else
	{
		memcpy(value, str, ULONG_SIZE);
	}
}

void AFFYBAR_READ_FLOAT(FILE *fpIn, float *value)
{
	char str[FLOAT_SIZE];
	char str2[FLOAT_SIZE];
	int ni;
	
	fread(str, sizeof(char), FLOAT_SIZE, fpIn);
	if(AFFY_IS_BIG_ENDIAN == 1)
	{
		for(ni=0; ni<FLOAT_SIZE; ni++)
			str2[ni] = str[FLOAT_SIZE-1-ni];
		memcpy(value, str2, FLOAT_SIZE);
	}
	else
	{
		memcpy(value, str, FLOAT_SIZE);
	}
}

void AFFYBAR_READ_VERSION(FILE *fpIn, float *value)
{
	char str[FLOAT_SIZE];
	char str2[FLOAT_SIZE];
	int x,y,ni;
	float z;
	
	fread(str, sizeof(char), FLOAT_SIZE, fpIn);
	
	/* for big endian */
	x = *((int *)(str));
	z = *((float*)&x);
	x = (int)z;
	x = ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24));
	if( (x==1) || (x==2) || (x==3) )
	{
		AFFY_IS_BIG_ENDIAN = 1;
		*value = (float)x;	
	}
	else
	{
		x = *((int *)(str));
		x = ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24));
		y = (int)(*(float*)(&x));
		y = ((((y) & 0xff000000) >> 24) | (((y) & 0x00ff0000) >>  8) | (((y) & 0x0000ff00) <<  8) | (((y) & 0x000000ff) << 24));
		
		if( (y==1) || (y==2) || (y==3) )
		{
			AFFY_IS_BIG_ENDIAN = 0;
			*value = (float)y;
		}

		else
		{
			for(ni=0; ni<FLOAT_SIZE; ni++)
				str2[ni] = str[FLOAT_SIZE-1-ni];
			z = *((float *)str2);
			y = (int)z;
			
			if( (y==1) || (y==2) || (y==3) )
			{
				AFFY_IS_BIG_ENDIAN = 1;
				*value = (float)y;
			}

			else
			{
				z = *((float *)str);
				y = (int)z;
				
				if( (y==1) || (y==2) || (y==3) )
				{
					AFFY_IS_BIG_ENDIAN = 0;
					*value = (float)y;
				}
				else
				{
					printf("Error: AFFYBAR_READ_VERSION, cannot read verion information of *.bpmap file!\n");
					exit(EXIT_FAILURE);
				}
			}
		}
	}

	printf("AFFY_IS_BIG_ENDIAN = %d\n", AFFY_IS_BIG_ENDIAN);
}
