/* ----------------------------------------------------------------------- */
/*  AffyLib.h : interface of the affymetrix library                        */
/*  Author : Ji HongKai ; Time: 2004.08                                    */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*                              Declaration                                */
/* ----------------------------------------------------------------------- */
#define MAX_MATCH_GAP 10000
#define TILE_PROBE_LEN 26
#define PROBE_BPMAP_LEN 7
#define DNA_BASE_A     0
#define DNA_BASE_C     1
#define DNA_BASE_G     2
#define DNA_BASE_T     3
#define DNA_BASE_OTHER 4
#define LOCAL_REPEAT_RANGE 1000

#define FLOAT_SIZE sizeof(float)
#define INT_SIZE sizeof(int)
#define ULONG_SIZE sizeof(unsigned long)
#define LONG_SIZE sizeof(long)
#define SHORT_SIZE sizeof(short)
#define USHORT_SIZE sizeof(unsigned short)
#define CHAR_SIZE sizeof(char)
#define UCHAR_SIZE sizeof(unsigned char)

/* ----------------------------------------------------------------------- */
/*                              Structures                                 */
/* ----------------------------------------------------------------------- */
struct tagAffyGenomeAlign
{
	/* chromosome */
	char strChr[30];
	int nChr;
	/* start */
	int nStart;
	/* end */
	int nEnd;
	/* rc */
	char chRc;
	/* identity */
	double dIdentity;
	/* physical map */
	char strPhyMap[30];
	/* refseq id */
	char strRefId[30];

	/* next */
	struct tagAffyGenomeAlign *pNext;
};


struct tagAffyBpMapUnit
{
	/* X coordinate of PM */
	unsigned long nPMX;
	/* Y coordinate of PM */
	unsigned long nPMY;
	/* X coordinate of MM */
	unsigned long nMMX;
	/* Y coordinate of MM */
	unsigned long nMMY;
	/* length of PM or MM probe */
	unsigned char bProbeLen;
	/* probe sequence */
	char strProbeSeq[TILE_PROBE_LEN];
	/* match score */
	float fMatchScore;
	/* position in the sequence */
	unsigned long nPos;
	/* strand */
	unsigned char bStrand;

	/* local repeat times */
	int nRepeatNum;
	/* same postion map times */
	int nDepthNum;

	/* next */
	struct tagAffyBpMapUnit *pNext;
};

/* ----------------------------------------------------------------------- */ 
/*                                Functions                                */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */ 
/*  Affy_CSVANNOT_To_GenomeAlign_200506()                                  */
/*  convert affy csv annotation file to genome mapping coordinates.        */
/* ----------------------------------------------------------------------- */ 
int Affy_CSVANNOT_To_GenomeAlign_200506(char strInFile[], char strOutFile[], char strSpecies[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_CSVANNOT_To_Reduced_200408()                                      */
/*  convert affy csv annotation file to reduced annotation                 */
/* ----------------------------------------------------------------------- */ 
int Affy_CSVANNOT_To_Reduced_200408(char strInFile[], char strOutFile[], char strSpecies[]);

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeAlignCreate()                                                */
/*  create genome alignment element                                        */
/* ----------------------------------------------------------------------- */ 
struct tagAffyGenomeAlign *AffyGenomeAlignCreate();

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeAlignDestroy()                                               */
/*  create genome alignment element                                        */
/* ----------------------------------------------------------------------- */ 
int AffyGenomeAlignDestroy(struct tagAffyGenomeAlign *pGA);

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeAlignInit()                                                  */
/*  init alignment element                                                 */
/* ----------------------------------------------------------------------- */ 
int AffyGenomeAlignInit(struct tagAffyGenomeAlign *pGA, char strParam[], char strSpecies[]);

/* ----------------------------------------------------------------------- */ 
/*  AffyGenomeTransInit()                                                  */
/*  init alignment element                                                 */
/* ----------------------------------------------------------------------- */ 
int AffyGenomeTransInit(struct tagAffyGenomeAlign *pGA, char strParam[], char strSpecies[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_PickRandomControlFromCSV()                                        */
/*  pick random control probesets from affymetrix's CSV annotation file.   */
/* ----------------------------------------------------------------------- */ 
int Affy_PickRandomControlFromCSV(char strInFile[], char strOutFile[], int nControlNum);

/* ----------------------------------------------------------------------- */ 
/*  Affy_MapProbesetToRefSeq_1()                                           */
/*  map probeset to refseq.                                                */
/* ----------------------------------------------------------------------- */ 
int Affy_LinkProbesetToAnnotation(char strProbesetFile[], char strAnnotationFile[], 
								  char strOutputFile[], int nClassId);


/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBar_Intensity()                                               */
/*  load intensity data from *.bar file                                    */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBar_Intensity(char strBarFile[], char strOutPath[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBar_Intensity_Group()                                         */
/*  load a group of intensity data from *.bar file                         */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBar_Intensity_Group(char strDataPath[], char strBarFileList[], char strOutPath[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_BPMAPFilter_GTrans()                                              */
/*  Filter out repeat and bad probes with bpmap informaiton.               */
/* ----------------------------------------------------------------------- */ 
int Affy_BPMAPFilter_GTrans(char strInFile[], char strRefFile[], char strOutFile[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_BPMAPFilter_ORI()                                                 */
/*  Filter out repeat and bad probes with bpmap informaiton.               */
/* ----------------------------------------------------------------------- */ 
int Affy_BPMAPFilter_ORI(char strInFile[], char strRefFile[], char strOutFile[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_BPMAPFilter_DATA()                                                 */
/*  Filter out repeat and bad probes with bpmap informaiton.               */
/* ----------------------------------------------------------------------- */ 
int Affy_BPMAPFilter_DATA(char strInFile[], char strRefFile[], char strOutFile[]);


/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBPMAP()                                                       */
/*  load bpmap data from *.bar file.                                       */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBPMAP(char strInFile[], char strOutFile[]);

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBPMAP_TileMap()                                               */
/*  load bpmap data from *.bar file, tilemap format                        */
/* ----------------------------------------------------------------------- */ 
int Affy_LoadBPMAP_TileMap(char strInFile[], char strPosFile[], 
						   char strMaskFile[], int *pnTotalProbeNum);


/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitCreate()                                                  */
/*  create bpmap unit.                                                     */
/* ----------------------------------------------------------------------- */ 
struct tagAffyBpMapUnit *AffyBpMapUnitCreate();

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitDestroy()                                                 */
/*  delete bpmap unit.                                                     */
/* ----------------------------------------------------------------------- */ 
int AffyBpMapUnitDestroy(struct tagAffyBpMapUnit *pUnit);

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitLoad()                                                    */
/*  load bpmap unit.                                                       */
/* ----------------------------------------------------------------------- */ 
int AffyBpMapUnitLoad(struct tagAffyBpMapUnit *pUnit, FILE *fpIn);

/* ----------------------------------------------------------------------- */ 
/*  AffyBpMapUnitLoad_v3()                                                 */
/*  load bpmap unit.                                                       */
/* ----------------------------------------------------------------------- */ 
int AffyBpMapUnitLoad_v3(struct tagAffyBpMapUnit *pUnit, FILE *fpIn, int nPMType);

/* ----------------------------------------------------------------------- */ 
/*  Functions for loading affy binary files.                               */
/* ----------------------------------------------------------------------- */ 
void AFFYBAR_READ_INT(FILE *fpIn, int *value);
void AFFYBAR_READ_ULONG(FILE *fpIn, unsigned long *value);
void AFFYBAR_READ_FLOAT(FILE *fpIn, float *value);
void AFFYBAR_READ_VERSION(FILE *fpIn, float *value);


