/* ----------------------------------------------------------------------- */
/*  GenomeLib.c : implementation of the genome library                     */
/*  Author : Ji HongKai ; Time: 2004.08                                    */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"

#include "MathLib.h"
#include "MatrixLib.h"
#include "StringLib.h"
#include "SequenceLib.h"
#include "GenomeLib.h"



/* ----------------------------------------------------------------------- */ 
/*  Genome_Index_To_ChromosomeName:                                        */
/*  convert numeric chromosome id to chromsome name                        */
/* ----------------------------------------------------------------------- */ 
char *Genome_Index_To_ChromosomeName(char *strChrName, char strSpecies[], int nChr)
{
	/* define */
	char strLine[LINE_LENGTH];
	char strChr[LINE_LENGTH];

	/* convert */
	strcpy(strLine, strSpecies);
	StrMakeUpper(strLine);

	/* human */
	if(strcmp(strLine, "HUMAN") == 0)
	{
		if( (nChr >= 1) && (nChr <= 22) )
		{
			sprintf(strChr, "chr%d", nChr);
		}
		else if(nChr == 23)
		{ 
			strcpy(strChr, "chrX");
		}
		else if(nChr == 24)
		{
			strcpy(strChr, "chrY");
		}
		else
		{
			strcpy(strChr, "unplaced");
		}
	}

	/* mouse */
	else if(strcmp(strLine, "MOUSE") == 0)
	{
		if( (nChr >= 1) && (nChr <= 19) )
		{
			sprintf(strChr, "chr%d", nChr);
		}
		else if(nChr == 20)
		{ 
			strcpy(strChr, "chrX");
		}
		else if(nChr == 21)
		{
			strcpy(strChr, "chrY");
		}
		else
		{
			strcpy(strChr, "unplaced");
		}
	}

	/* dog */
	else if(strcmp(strLine, "DOG") == 0)
	{
		if( (nChr >= 1) && (nChr <= 38) )
		{
			sprintf(strChr, "chr%d", nChr);
		}
		else if(nChr == 39)
		{ 
			strcpy(strChr, "chrX");
		}
		else if(nChr == 40)
		{
			strcpy(strChr, "chrY");
		}
		else
		{
			strcpy(strChr, "unplaced");
		}
	}

	/* fly */
	else if(strcmp(strLine, "D_MELANOGASTER") == 0)
	{
		if( nChr == 1 )
		{
			sprintf(strChr, "chr2L");
		}
		else if( nChr == 2 )
		{
			sprintf(strChr, "chr2R");
		}
		else if( nChr == 3 )
		{
			sprintf(strChr, "chr2h");
		}
		else if( nChr == 4 )
		{
			sprintf(strChr, "chr3L");
		}
		else if( nChr == 5 )
		{
			sprintf(strChr, "chr3R");
		}
		else if( nChr == 6 )
		{
			sprintf(strChr, "chr3h");
		}
		else if( nChr == 7 )
		{
			sprintf(strChr, "chr4");
		}
		else if( nChr == 8 )
		{
			sprintf(strChr, "chr4h");
		}
		else if( nChr == 9 )
		{
			sprintf(strChr, "chrX");
		}
		else if( nChr == 10 )
		{
			sprintf(strChr, "chrXh");
		}
		else if( nChr == 11 )
		{
			sprintf(strChr, "chrYh");
		}
		else
		{
			strcpy(strChr, "unplaced");
		}
	}

	/* others */
	else
	{
		strcpy(strChr, "unplaced");
	}
	
	strcpy(strChrName, strChr);

	/* return */
	return strChrName;
}

/* ----------------------------------------------------------------------- */ 
/*  Genome_ChromosomeName_To_Index:                                        */
/*  convert chromsome name to numeric chromosome id.                       */
/* ----------------------------------------------------------------------- */ 
int Genome_ChromosomeName_To_Index(char strChrName[], char strSpecies[])
{
	/* define */
	int nChrid;
	char strLine[LINE_LENGTH];
	char strChr[LINE_LENGTH];
	char *chp;

	/* convert */
	strcpy(strLine, strSpecies);
	StrMakeUpper(strLine);
	strcpy(strChr, strChrName);
	StrMakeUpper(strChr);
	StrTrimLeft(strChr);
	StrTrimRight(strChr);

	/* human */
	if(strcmp(strLine, "HUMAN") == 0)
	{
		if(strstr(strChr, "CHR") != strChr)
			return -1;
	
		if( (strstr(strChr, "_") != NULL) || (strstr(strChr, "RANDOM") != NULL) 
			|| (strstr(strChr, "UNPLACED") != NULL) )
			return -1;

		chp = strChr+3;
		if(*chp == 'X')
		{
			nChrid = 23;
		}
		else if(*chp == 'Y')
		{
			nChrid = 24;
		}
		else if((*chp >= '1') && (*chp <= '9'))
		{
			nChrid = atoi(chp);
		}
		else
		{
			nChrid = -1;
		}
	}

	/* mouse */
	else if(strcmp(strLine, "MOUSE") == 0)
	{
		if(strstr(strChr, "CHR") != strChr)
			return -1;
	
		if( (strstr(strChr, "_") != NULL) || (strstr(strChr, "RANDOM") != NULL) 
			|| (strstr(strChr, "UNPLACED") != NULL) )
			return -1;

		chp = strChr+3;
		if(*chp == 'X')
		{
			nChrid = 20;
		}
		else if(*chp == 'Y')
		{
			nChrid = 21;
		}
		else if((*chp >= '1') && (*chp <= '9'))
		{
			nChrid = atoi(chp);
		}
		else
		{
			nChrid = -1;
		}
	}


	/* dog */
	else if(strcmp(strLine, "DOG") == 0)
	{
		if(strstr(strChr, "CHR") != strChr)
			return -1;
	
		if( (strstr(strChr, "_") != NULL) || (strstr(strChr, "RANDOM") != NULL) 
			|| (strstr(strChr, "UNPLACED") != NULL) )
			return -1;

		chp = strChr+3;
		if(*chp == 'X')
		{
			nChrid = 39;
		}
		else if(*chp == 'Y')
		{
			nChrid = 40;
		}
		else if((*chp >= '1') && (*chp <= '9'))
		{
			nChrid = atoi(chp);
		}
		else
		{
			nChrid = -1;
		}
	}

	/* fly */
	else if(strcmp(strLine, "D_MELANOGASTER") == 0)
	{
		if(strstr(strChr, "CHR") != strChr)
			return -1;
	
		if( (strstr(strChr, "_") != NULL) || (strstr(strChr, "RANDOM") != NULL) 
			|| (strstr(strChr, "UNPLACED") != NULL) )
			return -1;

		if(strcmp(strChr, "CHR2L") == 0)
		{
			nChrid = 1;
		}
		else if(strcmp(strChr, "CHR2R") == 0)
		{
			nChrid = 2;
		}
		else if(strcmp(strChr, "CHR2H") == 0)
		{
			nChrid = 3;
		}
		else if(strcmp(strChr, "CHR3L") == 0)
		{
			nChrid = 4;
		}
		else if(strcmp(strChr, "CHR3R") == 0)
		{
			nChrid = 5;
		}
		else if(strcmp(strChr, "CHR3H") == 0)
		{
			nChrid = 6;
		}
		else if(strcmp(strChr, "CHR4") == 0)
		{
			nChrid = 7;
		}
		else if(strcmp(strChr, "CHR4H") == 0)
		{
			nChrid = 8;
		}
		else if(strcmp(strChr, "CHRX") == 0)
		{
			nChrid = 9;
		}
		else if(strcmp(strChr, "CHRXH") == 0)
		{
			nChrid = 10;
		}
		else if(strcmp(strChr, "CHRYH") == 0)
		{
			nChrid = 11;
		}
		else
		{
			nChrid = -1;
		}

	}

	/* others */
	else
	{
		nChrid = -1;
	}
	
	
	/* return */
	return nChrid;
}


