/* ----------------------------------------------------------------------- */
/*  MathLib.c : implementation of the mathematics library                  */
/*  Author : Ji HongKai ; Time: 2004.07                                    */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"
#include "time.h"

#include "MathLib.h"

/* ----------------------------------------------------------------------- */
/*                            lnGamma Function                             */
/* Return ln(gamma(dx)).                                                   */
/* ----------------------------------------------------------------------- */
double gammaln(double dx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,
						-86.50532032941677,
						24.01409824083091,
						-1.231739572450155,
						0.1208650973866179e-2,
						-0.5395239384953e-5};
	int j;

	if(dx<0.0)
	{
		printf("Error: gammaln(dx), dx must be greater than 0!\n");
		exit(EXIT_FAILURE);
	}
	y = dx;
	x = dx;
	tmp = x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser = 1.000000000190015;
	for(j=0; j<=5; j++)
		ser += cof[j]/++y;

	return -tmp+log(2.5066282746310005*ser/x);
}

/* ----------------------------------------------------------------------- */
/*                              Gamma Function                             */
/* Return gamma(dx).                                                       */
/* ----------------------------------------------------------------------- */
double gamma(double dx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,
						-86.50532032941677,
						24.01409824083091,
						-1.231739572450155,
						0.1208650973866179e-2,
						-0.5395239384953e-5};
	int j;

	if(dx<0.0)
	{
		printf("Error: gammaln(dx), dx must be greater than 0!\n");
		exit(EXIT_FAILURE);
	}
	y = dx;
	x = dx;
	tmp = x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser = 1.000000000190015;
	for(j=0; j<=5; j++)
		ser += cof[j]/++y;

	return exp(-tmp+log(2.5066282746310005*ser/x));
}

/* ----------------------------------------------------------------------- */
/*                           Factorial Function                            */
/* Return nx! = 1*2*...*nx.                                                */
/* ----------------------------------------------------------------------- */
double factorial(int nx)
{
	static int ntop=4;
	static double a[33] = {1.0,1.0,2.0,6.0,24.0};
	int j;

	if(nx<0)
	{
		printf("Error: factorial(nx), nx must be nonnegative!\n");
		exit(EXIT_FAILURE);
	}
	if(nx>32)
		return gamma(nx+1.0);

	while(ntop<nx)
	{
		j=ntop++;
		a[ntop]=a[j]*ntop;
	}
	return a[nx];
}

/* ----------------------------------------------------------------------- */
/*                               Beta Function                             */
/* Return beta(dx).                                                        */
/* ----------------------------------------------------------------------- */
double beta(double da, double db)
{
	return exp(gammaln(da)+gammaln(db)-gammaln(da+db));
}